from pathlib import Path
import pandas as pd
import numpy as np
import re
import csv # for pd_load quote-chars
from collections import Counter
import logging
from pandas.io.parsers import ParserError

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel( logging.INFO )

__all__ = ['read_geo', 'detect_header_pattern']

''' circular imports problems --- https://stackabuse.com/python-circular-imports/
try:
    # first: try to map the canonical version here
    import methylprep
    read_geo = methylprep.read_geo
except ImportError as error:
    # if user doesn't have methylprep installed for the canonical version of this function, import this copy below
'''

def read_geo_v1(filepath, verbose=False, debug=False):
    """VERSION 1.0 DEPRECATED June 2020 for v3, called "read_geo"
    Use to load preprocessed GEO data into methylcheck. Attempts to find the sample beta/M_values
    in the CSV/TXT/XLSX file and turn it into a clean dataframe, with probe ids in the index/rows.


    - reads a downloaded file, either in csv, xlsx, pickle, txt
    - looks for /d_RxxCxx patterned headings and an probe index
    - sets index in df to probes
    - sets columns to sample names
    - forces probe values to be floats, if strings/mixed
    - if filename has 'intensit' or 'signal' in it, this converts to betas and saves
      even if filename doesn't match, if columns have Methylated in them, it will convert and save
    - detect multi-line headers and adjusts dataframe columns accordingly
    - returns the usable dataframe

TODO:
    - handle files with .Signal_A and .Signal_B instead of Meth/Unmeth
    - handle processed files with sample_XX

notes:
    this makes inferences based on strings in the filename, and based on the column names.
    """
    this = Path(filepath)

    if '.csv' in this.suffixes:
        raw = pd.read_csv(this)
    elif '.xlsx' in this.suffixes:
        raw = pd.read_excel(this)
    elif '.pkl' in this.suffixes:
        raw = pd.read_pickle(this)
        return raw
    elif '.txt' in this.suffixes:
        raw = pd.read_csv(this, sep='\t')
        if raw.shape[1] == 1: # pandas doesn't handle \r\n two char line terminators, but seems to handle windows default if unspecified.
            raw = pd.read_csv(this, sep='\t', lineterminator='\r') # leaves \n in values of first column, but loads
            # lineterminator='\r\n')
            # or use codecs first to load and parse text file before dataframing...
    else:
        print(f'ERROR: this file type (){this.suffix}) is not supported')
        return

    # next, see if betas are present of do we need to calculate them?
    test = raw.iloc[0:100]
    unmeth = False
    if 'intensit' in str(this.name).lower() or 'signal' in str(this.name).lower(): # signal intensities
        unmeth = True # need to calculate beta from unmeth/meth columns
        print('Expecting raw meth/unmeth probe data')
    else:
        #meth_pattern_v1 = re.compile(r'.*[_ \.]Methylated[_ \.]', re.I)
        meth_pattern = re.compile(r'.*[_ \.]?(Un)?methylated[_ \.]?', re.I)
        meth_cols = len([col for col in test.columns if re.match(meth_pattern, col)])
        if meth_cols > 0:
            unmeth = True
            # this should work below, so that even if betas are present, it will use betas first, then fall back to meth/unmeth

    def calculate_beta_value(methylated_series, unmethylated_series, offset=100):
        """ borrowed from methylprep.processing.postprocess.py """
        methylated = max(methylated_series, 0)
        unmethylated = max(unmethylated_series, 0)
        total_intensity = methylated + unmethylated + offset
        intensity_ratio = methylated / total_intensity
        return intensity_ratio

    # look for probe names in values (of first 100 responses)
    index_name = None
    multiline_header = False
    sample_pattern = re.compile(r'\w?\d+_R\d{2}C\d{2}$') # $ ensures column ends with the regex part
    sample_pattern_loose = re.compile(r'\w?\d+_R\d{2}C\d{2}.*beta', re.I)
    probe_pattern = re.compile(r'(cg|rs|ch\.\d+\.|ch\.X\.|ch\.Y\.)\d+')
    samples = []
    for col in test.columns:
        probes = [i for i in test[col] if isinstance(i,str) and re.match(probe_pattern,i)] #re.match('cg\d+',i)]
        if len(probes) == len(test):
            index_name = col
            if verbose:
                print(f"Found probe names in `{col}` column and setting as index.")
        elif len(probes)/len(test) > 0.8:
            index_name = col
            multiline_header = True
            break

        if re.match(sample_pattern, col):
            samples.append(col)

    if multiline_header: # start over with new column names
        try:
            start_index = len(test) - len(probes) - 1
            # recast without header, starting at row before first probe
            new_column_names = pd.Series(list(raw.iloc[start_index])).replace(np.nan, 'No label')
            probe_list = raw[index_name].iloc[start_index + 1:]
            probe_list = probe_list.rename(raw[index_name].iloc[start_index + 1])
            bad_probe_list = [probe for probe in probe_list if not re.match(probe_pattern, probe)] # not probe.startswith('cg')]
            if bad_probe_list != []:
                print(f'ERROR reading probes with multiline header: {bad_probe_list[:200]}')
                return
            raw = raw.iloc[start_index + 1:]
            raw.columns = new_column_names
            test = raw.iloc[0:100]
            samples = []
            for col in test.columns:
                if re.match(sample_pattern, col):
                    samples.append(col)
            # raw has changed.
            out_df = pd.DataFrame(index=probe_list)
        except Exception as e:
            print("ERROR: Unable to parse the multi-line header in this file. If you manually edit the file headers to ensure the sample intensities unclude 'Methylated' and 'Unmethylated' in column names, it might work on retry.")
            return
    else:
        out_df = pd.DataFrame(index=raw[index_name]) # only used with unmethylated data sets

    if samples == []:
        # in some cases there are multiple columns matching sample_ids, and we want the 'beta' one
        for col in test.columns:
            if re.match(sample_pattern_loose, col):
                samples.append(col)

        # or we need TWO columns per sample and we calculate 'beta'.
        if samples == [] and unmeth:
            vectorized_beta_func = np.vectorize(calculate_beta_value)
            unmeth_samples = []
            meth_samples = []
            #unmeth_pattern_v1 = re.compile(r'.*[_ \.]Unmethylated[_ \.].*', re.I)
            #meth_pattern_v1 = re.compile(r'.*[_ \.]Methylated[_ \.].*', re.I)
            unmeth_pattern = re.compile(r'.*[_ \.]?Unmethylated[_ \.]?', re.I)
            meth_pattern = re.compile(r'.*[_ \.]?(?<!Un)Methylated[_ \.]?', re.I)
            for col in test.columns:
                if re.match(unmeth_pattern, col):
                    unmeth_samples.append(col)
                    if debug:
                        print(col)
                if re.match(meth_pattern, col):
                    meth_samples.append(col)
                    if debug:
                        print(col)
            if unmeth_samples != [] and meth_samples != [] and len(unmeth_samples) == len(meth_samples):
                # next: just need to match these up. they should be same if we drop the meth/unmeth part
                if verbose:
                    print(f"{len(unmeth_samples)} Samples with Methylated/Unmethylated probes intensities found. Calculating Beta Values.")
                linked = []
                for col in unmeth_samples:
                    test_name = col.replace('Unmethylated','Methylated')
                    if test_name in meth_samples:
                        linked.append([col, test_name])
                # Here, we calculate betas for full raw data frame
                for col_u, col_m in linked:
                    col_name = col_u.replace('Unmethylated','').replace('Signal','').strip()
                    unmeth_series = raw[col_u]
                    meth_series = raw[col_m]
                    betas = vectorized_beta_func(meth_series, unmeth_series)
                    try:
                        out_df[col_name] = betas
                        samples.append(col_name)
                    except Exception as e:
                        print('ERROR', col_name, len(betas), out_df.shape, e)
            elif unmeth:
                print(f"File appears to contain probe intensities, but the column names don't match up for samples, so can't calculate beta values.")

        if samples != [] and verbose and not unmeth:
            print(f"Found {len(samples)} samples on second pass, apparently beta values with a non-standard sample_naming convention.")
        elif samples != [] and verbose and unmeth:
            pass
        elif samples == []:
            # no samples matched, so show the columns instead
            print(f"No samples found. Here are some column names:")
            print(list(test.columns)[:20])
            return

    if index_name == None:
        print("Error: probe names not found in any columns")
        return
    if unmeth and samples != [] and out_df.shape[1] > 1:
        # column names are being merged and remapped here as betas
        df = out_df # index is already set
    elif multiline_header:
        df = raw.loc[:, samples]
        df.index = probe_list
    else:
        df = raw[[index_name] + samples]
        df = df.set_index(index_name)

    # finally, force probe values to be floats
    num_converted = 0
    for col in df.columns:
        if df[col].dtype.kind != 'f' and df[col].dtype.kind == 'O':
            # convert string to float
            try:
                #df[col] = df[col].astype('float16')
                # use THIS when mix of numbers and strings
                df[col] = pd.to_numeric(df[col], errors='coerce')
                num_converted += 1
            except:
                print('error')
                df = df.drop(columns=[col])
    if verbose:
        if num_converted > 0:
            print(f"Converted {num_converted} samples from string to float16.")
        print(f"Found {len(samples)} samples and dropped {len(raw.columns) - len(samples)} meta data columns.")
    return df


def read_geo(filepath, verbose=False, debug=False, as_beta=True, column_pattern=None, test_only=False, rename_probe_column=True, decimals=3):
    """Use to load preprocessed GEO data into methylcheck. Attempts to find the sample beta/M_values
    in the CSV/TXT/XLSX file and turn it into a clean dataframe, with probe ids in the index/rows.
    Version 3 (introduced June 2020)

    - reads a downloaded file, either in csv, xlsx, pickle, txt
    - looks for /d_RxxCxx patterned headings and an probe index
    - sets index in df to probes
    - sets columns to sample names
    - forces probe values to be floats, if strings/mixed
    - if filename has 'intensit' or 'signal' in it, this converts to betas and saves
      even if filename doesn't match, if columns have Methylated in them, it will convert and save
    - detect multi-line headers and adjusts dataframe columns accordingly
    - returns the usable dataframe

    as_beta == True -- converts meth/unmeth into a df of sample betas.
    column_pattern=None (Sample21 | Sample_21 | Sample 21) -- some string of characters that precedes the number part of each sample in the columns of the file to be ingested.

FIXED:
    [x] handle files with .Signal_A and .Signal_B instead of Meth/Unmeth
    [x] BUG: can't parse matrix_... files if uses underscores instead of spaces around sample numbers, or where sampleXXX has no separator.
    [x] handle processed files with sample_XX
    [x] returns IlmnID as index/probe column, unless 'rename_probe_column' == False
    [x] pass in sample_column names from header parser so that logic is in one place
        (makes the output much larger, so add kwarg to exclude this)
    [x] demicals (default 3) -- round all probe beta/intensity/p values returned to this number of decimal places.
    [x] bug: can only recognize beta samples if 'sample' in column name, or sentrix_id pattern matches columns.
        need to expand this to handle arbitrary sample naming styles (limited to one column per sample patterns)
TODO:
    [-] BUG: meth_unmeth_pval works `as_beta` but not returning full data yet
    [-] multiline header not working with all files yet.
    [-] _family GSM123456-tbl-1.txt files not detected yet

notes:
    this makes inferences based on strings in the filename, and based on the column names.
    """

    #1. load and read the file, regardless of type and CSV delimiter choice.
    this = Path(filepath)
    kwargs = {'nrows':200} if test_only else {}

    raw = pd_load(filepath, **kwargs)
    if not isinstance(raw, pd.DataFrame):
        LOGGER.error(f"Did not detect a file: {type(raw)} aborting")
        return
    if debug: LOGGER.info(f"{filepath} loaded.")

    def calculate_beta_value(methylated_series, unmethylated_series, offset=100):
        """ borrowed from methylprep.processing.postprocess.py """
        methylated = np.clip(methylated_series, 0, None)
        unmethylated = np.clip(unmethylated_series, 0, None)

        total_intensity = methylated + unmethylated + offset
        intensity_ratio = methylated / total_intensity
        return intensity_ratio

    #if as_beta:
    def convert_meth_to_beta(raw, meta, rename_probe_column=True, decimals=3, verbose=False):
        #if (meta['column_pattern'] == 'meth_unmeth_pval' and
        #    meta['sample_columns_meth_unmeth_count'] > 0 and
        #    'sample_numbers_range' != []):
        # beta = methylated / total intensity (meth + unmeth + 100)
        # use sample_column_names and sample_numbers_list to find each set of columns
        out_df = pd.DataFrame(index=raw[meta['probe_column_name']])
        probe_name_msg = meta['probe_column_name']
        if rename_probe_column:
            out_df.index.name = 'IlmnID' # standard for methyl-suite, though ID_REF is common in GEO.
            probe_name_msg = f"{meta['probe_column_name']} --> IlmnID"

        if not meta.get('sample_names') and not meta['sample_names'].get('meth'):
            # when "sample" not in columns.
            LOGGER.error("ERROR: missing sample_names")
            pass
        if meta['sample_numbers_range'] and meta['sample_numbers_range'][1] <= meta['total_samples']:
            # if non-sequential sample numbers, cannot rely on this range.
            sample_min, sample_max = meta['sample_numbers_range']
        else:
            sample_min = 1
            sample_max = meta['total_samples']

        # JSON returns 1 to N (as human names) but index is 0 to N.
        for sample_number in range(sample_min -1, sample_max):
            # here need to get the corresponding parts of each sample triplet of columns.
            try:
                col_m = meta['sample_names']['meth'][sample_number]
                col_u = meta['sample_names']['unmeth'][sample_number]
                #col_pval = meta['sample_columns']['pval']
            except Exception as e:
                LOGGER.error(f"ERROR {e} {meta['sample_names']['meth'][sample_number]} {list(meta['sample_names']['unmeth'].columns)[sample_number]}")
                continue

            unmeth_series = raw[col_u]
            meth_series = raw[col_m]
            betas = calculate_beta_value(meth_series, unmeth_series)

            # try to lookup and retain any unique part of column names here
            col_name = f"Sample_{sample_number + 1}"
            if isinstance(meta['sample_names_stems'], list):
                # assume order of sample_names_stems matches order of sample names
                try:
                    this_stem = meta['sample_names_stems'][sample_number]
                    if this_stem not in out_df.columns:
                        col_name = this_stem
                except IndexError:
                    if debug: LOGGER.error(f"ERROR: unable to assign Sample {sample_number} using original column stem.")

                """ This code was trying to match samples up using regex/difflib but it was unreliable.
                this_stem = [stem for stem in meta['sample_names_stems'] if stem in col_m]
                if len(this_stem) > 0:
                    if len(this_stem) > 1:
                        # difflib to separatet Sample 1 vs Sample 11
                        best_match = difflib.get_close_matches(col_m, this_stem, 1)
                        if best_match != [] and best_match[0] not in out_df.columns:
                            col_name = best_match[0]
                        elif best_match != []: # ensure unique in out_df, even if not a perfect transfer of labels.
                            col_name = f"{best_match[0]}_{sample_number}"
                        else:
                            if debug: LOGGER.info(f"WARNING: multiple similar sample names detected but none were a close match: {col_m} : {this_stem}")
                            col_name = f"Sample_{sample_number}"
                    elif len(this_stem) == 1 and this_stem[0] not in out_df.columns:
                        # only one match, and is unique.
                        col_name = this_stem[0]
                    else: # only one match, but already in out_df, so can't reuse.
                        col_name = f"{this_stem[0]}_Sample_{sample_number}"
                else:
                    col_name = f"Sample_{sample_number}"

            else:
                col_name = f"Sample_{sample_number}"
                """

            try:
                out_df[col_name] = betas
                sample_number += 1
            except Exception as e:
                LOGGER.error(f"ERROR {col_name} {len(betas)} {out_df.shape} {e}")
        out_df = out_df.round(decimals)

        beta_value_range = True if all([all(out_df[col_name].between(0,1)) == True for col_name in out_df.columns]) else False
        intensity_value_range = True if all([all(out_df[col_name].between(0,1000000)) == True for col_name in out_df.columns]) else False
        try:
            value_mean = round(sum(out_df.mean(axis=0))/len(out_df.columns),2)
        except:
            value_mean = 0
        if verbose:
            LOGGER.info(f"Returning {len(out_df.columns)} samples (mean: {value_mean}). Structure appears to be {meta['columns_per_sample']} columns per sample; raw data column pattern was '{meta['column_pattern']}'; probes in rows; and {probe_name_msg} as the probe names.")
        return out_df

    #2. test file structure
    # next, see if betas are present of do we need to calculate them?
    test = raw.iloc[0:100]
    meta = detect_header_pattern(test, filepath, return_sample_column_names=True)
    if meta['multiline_header'] and meta['multiline_header_rows'] > 0:
        if verbose: print("Reloading raw data, excluding header.")
        old_non_blank_col_count = len(raw.loc[:, ~raw.columns.str.contains('^Unnamed:')].columns) #~ before test counts non-blank columns
        kwargs['skiprows'] = meta['multiline_header_rows']
        raw = pd_load(filepath, **kwargs)
        new_non_blank_col_count = len(raw.loc[:, ~raw.columns.str.contains('^Unnamed:')].columns)
        if verbose: print(f"After ignoring multiline header: {old_non_blank_col_count} old columns are now {new_non_blank_col_count} new named columns.")
    if debug:
        LOGGER.info(f"file shape: {raw.shape}")
        concise = meta.copy()
        concise.pop('sample_names')
        from pprint import pprint
        pprint(concise)
        del concise

    #3. use header_meta to parse and return data as dataframe
    """
{'all_sample_columns': True,
 'column_pattern': 'meth_unmeth_pval',
 'columns_per_sample': 3,
 'fraction_sample pval meth unmeth signal intensity': (1.0, 0.33, 0.33, 0.33, 0.66, 0.0),
 'has_beta_values': False,
 'has_meth_unmeth_values': True,
 'has_p_values': True,
 'multiline_header': False,
 'multiline_rows': 0,
 'one_sample_beta': False,
 'probe_column_name': 'ID_REF',
 'sample_columns_meth_unmeth_count': 222,
 'sample_numbers_range': [1, 111], # or None
 'sequential_numbers': True,
 'total_samples': 111}"""

    if (meta['column_pattern'] == 'sample_beta_sentrix_id'):
        sample_columns = meta['sample_names']
        out_df = pd.DataFrame(data=raw[sample_columns])
        out_df.index=raw[meta['probe_column_name']]
        out_df = out_df.round(decimals) #{sample:demicals for sample in sample_columns})

        probe_name_msg = meta['probe_column_name']
        if rename_probe_column:
            out_df.index.name = 'IlmnID' # standard for methyl-suite, though ID_REF is common in GEO.
            probe_name_msg = f"{meta['probe_column_name']} --> IlmnID"

        beta_value_range = True if all([all(out_df[col_name].between(0,1)) == True for col_name in out_df.columns]) else False
        intensity_value_range = True if all([all(out_df[col_name].between(0,1000000)) == True for col_name in out_df.columns]) else False
        try:
            value_mean = round(sum(out_df.mean(axis=0))/len(out_df.columns),2)
        except:
            value_mean = 0
        if verbose and meta['columns_per_sample'] != 1 and beta_value_range:
            LOGGER.info(f"Returning raw data. Structure appears to be {meta['columns_per_sample']} columns per sample; numbered samples: {numbered_samples}; column_pattern: {column_pattern}; probes in rows; and {probe_name_msg} as the probe names.")
        elif verbose and beta_value_range:
            LOGGER.info(f"Returning raw data. Appears to be sample beta values in columns (mean: {value_mean}), probes in rows, and {probe_name_msg} as the probe names.")
        elif verbose and not beta_value_range and intensity_value_range and value_mean > 10:
            LOGGER.info(f"Returning raw data. Appears to be sample fluorescence intensity values in columns (mean: {int(value_mean)}), probes in rows, and {probe_name_msg} as the probe names.")
        elif verbose and not beta_value_range and intensity_value_range and value_mean > 10:
            LOGGER.info(f"Returning raw data of UNKNOWN type. Not beta values or fluorescence intensity values in columns. Mean probe value was {value_mean}. Probes are in rows, and {probe_name_msg} as the probe names.")
        return out_df

    elif ((meta['column_pattern'] == 'sample_beta_numbered') or

        (meta['all_sample_columns'] and meta['has_beta_values']) or
        (meta['sequential_numbers'] and meta['sample_columns_meth_unmeth_count'] == 0) or
        (meta['sample_numbers_range'] and meta['has_p_values'] is False) or
        (meta['all_sample_columns'] and meta['sample_numbers_range'] and meta['columns_per_sample'] == 1)
       ):
        sample_columns = [column for column in test.columns if 'sample' in column.lower()]
        out_df = pd.DataFrame(data=raw[sample_columns])
        out_df.index=raw[meta['probe_column_name']]
        out_df = out_df.round(decimals) #{sample:demicals for sample in sample_columns})
        probe_name_msg = meta['probe_column_name']
        if rename_probe_column:
            out_df.index.name = 'IlmnID' # standard for methyl-suite, though ID_REF is common in GEO.
            probe_name_msg = f"{meta['probe_column_name']} --> IlmnID"
        #if debug: LOGGER.info(f"DEBUG: out_df.columns: {out_df.columns} --- out_df.index {out_df.index.name} out_df.shape {out_df.shape}")

        beta_value_range = True if all([all(out_df[col_name].between(0,1)) == True for col_name in out_df.columns]) else False
        intensity_value_range = True if all([all(out_df[col_name].between(0,1000000)) == True for col_name in out_df.columns]) else False
        try:
            value_mean = round(sum(out_df.mean(axis=0))/len(out_df.columns),2)
        except:
            value_mean = 0
        if verbose and meta['columns_per_sample'] != 1 and beta_value_range:
            LOGGER.info(f"Returning raw data. Structure appears to be {meta['columns_per_sample']} columns per sample; {len(out_df.columns)} numbered samples; column_pattern: {column_pattern}; probes in rows; and {probe_name_msg} as the probe names.")
        elif verbose and beta_value_range:
            LOGGER.info(f"Returning raw data. Appears to contain {len(out_df.columns)} numbered samples; beta values in columns (mean: {value_mean}), probes in rows, and {probe_name_msg} as the probe names.")
        elif verbose and not beta_value_range and intensity_value_range and value_mean > 10:
            LOGGER.info(f"Returning raw data. Appears to be sample fluorescence intensity values in columns ({len(out_df.columns)} samples with mean: {int(value_mean)}), probes in rows, and {probe_name_msg} as the probe names.")
        elif verbose and not beta_value_range and intensity_value_range and value_mean > 10:
            LOGGER.info(f"Returning raw data of UNKNOWN type. Not beta values or fluorescence intensity values in columns. {len(out_df.columns)} samples with a mean probe value of {value_mean}. Probes are in rows, and {probe_name_msg} as the probe names.")

        return out_df

    elif meta['column_pattern'] == 'beta_intensity_pval':
        if verbose: LOGGER.info("returning raw data without processing. Column pattern was 'beta_intensity_pval'.")
        return raw
    elif meta['column_pattern'] == 'beta_pval_a_b':
        if as_beta and meta['sample_names'] != None:
            # only returning the beta part. Signal_A and Signal_B are the meth/unmeth parts.
            out_df = pd.DataFrame(data=raw[meta['sample_names']])
            out_df.index=raw[meta['probe_column_name']]
            out_df = out_df.round(decimals) #{sample:demicals for sample in sample_columns})
            probe_name_msg = meta['probe_column_name']
            if rename_probe_column:
                out_df.index.name = 'IlmnID' # standard for methyl-suite, though ID_REF is common in GEO.
                probe_name_msg = f"{meta['probe_column_name']} --> IlmnID"
            try:
                value_mean = round(sum(out_df.mean(axis=0))/len(out_df.columns),2)
            except:
                value_mean = 0
            if verbose:
                LOGGER.info(f"Returning beta values for {len(out_df.columns)} samples in columns (mean: {value_mean}), probes in rows, and {probe_name_msg} as the probe names.")
            return out_df
        else:
            if verbose: LOGGER.info("returning raw data without processing. Column pattern was 'beta_pval_a_b'.")
            return raw

    if meta['column_pattern'] == 'meth_unmeth':
        LOGGER.info("*** column_pattern was 'meth_unmeth' this has litterally never happened before. ***")
    if meta['column_pattern'] in ('meth_unmeth','meth_unmeth_pval') and as_beta:
        if debug: LOGGER.info("Converting meth and unmeth intensities to beta values.")
        return convert_meth_to_beta(raw, meta, rename_probe_column=rename_probe_column, verbose=verbose)
    if meta['column_pattern'] in ('meth_unmeth','meth_unmeth_pval') and not as_beta:
        if verbose: LOGGER.info(f"Returning raw data without processing. Column pattern was {meta['column_pattern']}.")
        return raw

    if meta['column_pattern'] is None:
        if debug: LOGGER.info("Returning raw data without processing. No file header column pattern was detected.")
        return raw
    raise Exception("Unable to identify file structure")


def detect_header_pattern(test, filename, return_sample_column_names=False):
    """test is a dataframe with first 100 rows of the data set, and all columns.
    makes all the assumptions easier to read in one place.

    betas
    non-normalized
    matrix_processed
    matrix_signal
    series_matrix
    methylated_signal_intensities and unmethylated_signal_intensities
    _family

    TODO: GSM12345-tbl-1.txt type files (in _family.tar.gz packages) are possible, but needs more work.
    TODO: combining two files with meth/unmeth values

    - numbered samples handled differently from sample_ids in columns
    - won't detect columns with no separators in strings
    """
    if test.shape[0] != 100:
        raise ValueError("test dataset must be exactly 100 rows")
    if test.shape[1] == 1:
        raise ValueError("this dataset has only one sample. it is likely that the columns were not parsed correctly.")

    seps = [' ', '_', '.', '-'] # for parsing columns. also try without any separators
    # index_names = ['IlmnID', 'ID_REF', 'illumina_id']

    # sample patterns
    sample_pattern = re.compile(r'\w?\d+_R\d{2}C\d{2}$') # $ ensures column ends with the regex part
    sample_pattern_loose = re.compile(r'\w?\d+_R\d{2}C\d{2}.*beta', re.I)
    samplelike_pattern = re.compile(r'.*(?:\w?\d+_R\d{2}C\d{2}|sample).*', re.I)
    probe_pattern = re.compile(r'(cg|rs|ch\.\d+\.|ch\.X\.|ch\.Y\.)\d+')
    #pval_pattern = re.compile(r'(.*)(?:\bPval\b|\.Pval|_Pval_|-Pval|Pval|\bDetection\bPval|_Detection_Pval|\._Detection\.Pval||\._Detection\bPval\b).*', re.I)
    pval_pattern = re.compile(r'(.*)(?:\bPval\b|\.Pval|_Pval_|-Pval|Pval).*', re.I)
    meth_pattern = re.compile(r'(.*)(?:\bmeth|\.meth|_meth|-meth|(?<!un)meth\w+).*', re.I)
    unmeth_pattern = re.compile(r'(.*)(?:\bunmeth|\.unmeth|_unmeth|-unmeth|unmeth\w+).*', re.I)
    intensity_pattern = re.compile(r'(.*)(?:\bintensity\b|\.intensity|_intensity|-intensity|intensity).*', re.I)
    signal_pattern = re.compile(r'(.*)(?:\bsignal\b|\.signal|_signal|-signal|signal).*', re.I)
    betalike_pattern = re.compile(r'(.*)(?:\Wbeta\W|\Wavg\Wbeta\W|\Wavg_beta|_avg_beta|\Wavgbeta).*', re.I)
    # use the last part to find any sample identifiers; then re.sub() replace them. then pass rest into the beta_converter function.
    residual_pattern = re.compile(r'(.*)(?:[\b\._-]Pval[\b\._-]|[\b\._-]Detection[\b\._-]Pval|[\b\._-]meth|(?<!un)meth\w+|\bunmeth|\.unmeth|_unmeth|-unmeth|unmeth\w+|\bintensity\b|\.intensity|_intensity|-intensity|intensity).*', re.I)
    file_gsm_txt_in_family_tar = re.compile(r'gsm\d+-tbl-1.txt') # all matches are lower() -- not used yet
    #meth_unmeth_pattern = re.compile(r'.*[_ \.]?(Un)?methylated[_ \.]?', re.I)

    # filename patterns
    this = str(Path(filename).name).lower()
    has_beta_values = True if ('matrix' in this and 'signal' not in this) else False #exceptions to this rule found.
    # for meth_unmeth, need to calculate beta from columns
    has_meth_unmeth_values = True if 'intensit' in this or 'signal' in this else False
    # no rule for this yet
    has_p_values = None
    # GSM1280914-tbl-1.txt contains probe names, one sample beta val column, and an optional p-val column with no header.
    one_sample_beta = True if this.startswith('gsm') and re.search(this, file_gsm_txt_in_family_tar) else False

    # column parts
    sample_columns = [column for column in test.columns if 'sample' in column.lower()]
    # -- next: test first 100 rows and calculate the faction of each column that matches the probe_pattern. list of floats.
    fraction_probelike = [len([row for row in test[column] if isinstance(row, str) and re.match(probe_pattern,row)])/len(test) for column in test.columns]
    fraction_pvallike = round(sum([(True if re.search(pval_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_meth = round(sum([(True if re.search(meth_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_unmeth = round(sum([(True if re.search(unmeth_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_intensity = round(sum([(True if re.search(intensity_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_signal = round(sum([(True if re.search(signal_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_beta = round(sum([(True if re.search(betalike_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    fraction_samplelike = round(sum([(True if re.search(samplelike_pattern,column) else False) for column in test.columns])/len(list(test.columns)),2)
    probe_column_position = fraction_probelike.index( max(fraction_probelike) )
    probe_column_name = list(test.columns)[probe_column_position] # used for index_name
    # grab residual column name parts to add to output column names (tricky)
    sample_column_residuals = [re.search(residual_pattern, column).groups()[0] for column in test.columns if re.search(residual_pattern, column)]
    # a trick to deal with 3-column samples that have similar names; retain any repeated names, if repeats exist.
    if sample_column_residuals != [] and Counter(sample_column_residuals).most_common(1)[0][1] > 1:
        sample_column_residuals = [column.strip() for column,freq in Counter(sample_column_residuals).most_common() if freq > 1]

    ## DETECT column repeating structure.
    # how many columns PER SAMPLE in structure?
    # meth or (un)meth appear anywhere in column name along with 'sample' anywhere
    sample_columns_meth_unmeth_count = sum([(True if (re.search(meth_pattern,column) or re.search(unmeth_pattern,column)) else False) for column in test.columns])
    #sum([('meth' in column.lower().replace('sample', '')) for column in sample_columns])

    # extract any "first" numbers found in column parts, where parts are strings separated any any logical separators.
    sample_id_columns = []
    sample_numbers_list = []
    sample_numbers_range = None
    sample_count = 0
    sequential_numbers = None # starting from 1 to xx
    avg_sample_repeats = 1
    failed_attempts = []
    for sep in seps:
        try:
            sample_numbers_list = [[part for part in column.split(sep) if re.match(r'\d+', part)][0] for column in sample_columns]
            if len(sample_numbers_list) > 0:
                sorted_sample_numbers_list = sorted([int(j) for j in list(set(sample_numbers_list))])
                sample_repeats = list(Counter([int(j) for j in list(sample_numbers_list)]).values())
                avg_sample_repeats = int(round(sum(sample_repeats)/len(sample_repeats)))
                #sequential_numbers = True if all([(i+1 == j) for i,j in list(enumerate(sorted_sample_numbers_list))]) else False
                sequential_numbers = True if all([(i+1 == j) for i,j in list(enumerate(sorted_sample_numbers_list))]) else False
                sample_count = len(sample_numbers_list)
                sample_numbers_range = [min(sorted_sample_numbers_list), max(sorted_sample_numbers_list)]
                break
        except Exception as e:
            failed_attempts.append(e)
            continue

    # detect if some part of columns are named like sample_ids
    if sample_numbers_list == []:
        for sep in seps:
            sample_id_columns = [column for column in test.columns if any([re.search(sample_pattern, part) for part in column.split(sep)])]
        sample_count = len(sample_id_columns)

    # tests / data attributes: pval, multiline, probes, sample/nonsample columns
    if max(fraction_probelike) < 0.75:
        LOGGER.warning(f"WARNING: Unable to identify the column with probe names ({max(fraction_probelike)})")
    multiline_rows = 0
    header_rows = 0
    if 0.75 <= max(fraction_probelike) < 1.00:
        multiline_rows = int(round(100*max(fraction_probelike)))
        header_rows = 100 - multiline_rows
        LOGGER.info(f"Multiline header detected with {header_rows} rows.")
    multiline_header = True if multiline_rows > 0 else False
    all_sample_columns = all([('sample' in column.lower() or 'ID_REF' in column) for column in list(set(test.columns))])
    has_p_values = True if fraction_pvallike > 0 else False

    # overall pattern logic
    column_pattern = None
    if (sample_numbers_list != [] or sample_id_columns != []):
        columns_per_sample = int(round(sample_count/len(set(sample_numbers_list or sample_id_columns))))
    elif (0.31 <= fraction_meth <= 0.34 and 0.31 <= fraction_unmeth <= 0.34 and 0.31 <= fraction_pvallike <= 0.34):
        columns_per_sample = 3
    else:
        columns_per_sample = 1

    if sample_numbers_list != [] and columns_per_sample != avg_sample_repeats:
        LOGGER.warning('WARNING: inconsistent columns per sample')
    if (sample_count > 0 and
        sample_numbers_list != [] and
        sample_columns_meth_unmeth_count == 0 and
        0.9 < columns_per_sample < 1.1):
        column_pattern = 'sample_beta_numbered'
        total_samples = len(sample_columns)

    elif (sample_count > 0 and
          sample_id_columns != [] and
          sample_columns_meth_unmeth_count == 0 and
          0.9 < columns_per_sample < 1.1):
        column_pattern = 'sample_beta_sentrix_id'
        total_samples = len(sample_columns)

    elif (sample_columns_meth_unmeth_count > 0 and
          sample_count > 0 and
          sample_columns_meth_unmeth_count > 0 and
          1.8 < len(sample_numbers_list) / sample_columns_meth_unmeth_count < 2.2):
        column_pattern = 'meth_unmeth'
        total_samples = int(round(sample_columns_meth_unmeth_count/2))

    elif (sample_columns_meth_unmeth_count > 0 and
          sample_count > 0 and
          sample_columns_meth_unmeth_count > 0 and
          1.4 < len(sample_numbers_list)/sample_columns_meth_unmeth_count < 1.6):
        column_pattern = 'meth_unmeth_pval'
        total_samples = int(round(sample_columns_meth_unmeth_count/2))

    elif (0.03 <= fraction_pvallike <= 0.36 and
          0.03 <= fraction_meth <= 0.36 and
          0.03 <= fraction_unmeth <= 0.36):
        column_pattern = 'meth_unmeth_pval'
        total_samples = int(round(len(test.columns)*fraction_pvallike))

    elif (0.30 <= fraction_pvallike <= 0.36 and sample_columns_meth_unmeth_count == 0):
        column_pattern = 'beta_intensity_pval'
        total_samples = int(round(len(test.columns)*fraction_pvallike))

    elif (0.45 <= fraction_signal <= 0.55 and
          0.22 <= fraction_pvallike <= 0.26):
        column_pattern = 'beta_pval_a_b' #SignalA ... SignalB
        total_samples = int(round(len(test.columns)*fraction_pvallike))

    else:
        column_pattern = None
        total_samples = 0

    # return_sample_column_names
    # -- depends on structure.
    # -- sample_columns will be a flat list for "sample"-like ones.
    # -- Or a dict of lists if meth_unmeth_pval.
    if return_sample_column_names:
        if fraction_meth > 0 and fraction_unmeth > 0 and column_pattern in ('meth_unmeth_pval', 'meth_unmeth'):
                sample_columns = {
                                  'meth': [column for column in test.columns if re.search(meth_pattern,column)],
                                  'unmeth': [column for column in test.columns if re.search(unmeth_pattern,column)],
                                  'pval': [column for column in test.columns if re.search(pval_pattern,column)],
                                 }
        elif sample_columns == [] and sample_id_columns != []:
            sample_columns = sample_id_columns
        elif sample_columns == [] and column_pattern == 'beta_intensity_pval':
            sample_columns = test.columns
        elif column_pattern == 'beta_pval_a_b': # and sample_columns == []
            if 0.2 <= fraction_beta <= 0.28:
                sample_columns = [column for column in test.columns if re.search(betalike_pattern,column)]
            else:
                LOGGER.warning("WARNING: test columns for beta_pval_a_b don't look right.")
                sample_columns = test.columns
    else:
        sample_columns = None
        sample_columns_residuals = None

    return {'has_beta_values': has_beta_values, #  or (beta_assumed and columns_per_sample == 1),
            'has_meth_unmeth_values': has_meth_unmeth_values,
            'has_p_values': has_p_values,
            'one_sample_beta': one_sample_beta,
            'sample_columns_meth_unmeth_count': sample_columns_meth_unmeth_count,
            'all_sample_columns': all_sample_columns,
            'multiline_header': multiline_header,
            'multiline_header_rows': header_rows,
            'column_pattern': column_pattern,
            'columns_per_sample': columns_per_sample,
            'sequential_numbers': sequential_numbers,
            'sample_numbers_range': sample_numbers_range,
            'total_samples': total_samples,
            'fraction_sample pval meth unmeth signal intensity beta': (fraction_samplelike, fraction_pvallike, fraction_meth, fraction_unmeth, fraction_signal, fraction_intensity, fraction_beta),
            'probe_column_name': probe_column_name,
            'sample_names': sample_columns,
            'sample_names_stems': sample_column_residuals, # unique parts of sample names
           }

def pd_load(filepath, **kwargs):
    """ helper function that reliably loads any GEO file, by testing for the separator that gives the most columns """
    this = Path(filepath)

    if this.suffix not in ('.xlsx', '.pkl'):
        # first, check that we're getting the max cols
        test_csv = pd.read_csv(this, nrows=100, skiprows=kwargs.get('skiprows',0))
        test_t = pd.read_csv(this, sep='\t', nrows=100, skiprows=kwargs.get('skiprows',0))
        test_space = pd.read_csv(this, sep=r',\s+', quoting=csv.QUOTE_ALL, engine='python', skiprows=kwargs.get('skiprows',0))

        params = [
            {'method':'auto', 'cols': test_csv.shape[1], 'kwargs': {}},
            {'method':'tab', 'cols': test_t.shape[1], 'kwargs': {'sep':'\t'}},
            {'method':'quoted', 'cols': test_space.shape[1], 'kwargs': {'sep':r',\s+', 'quoting':csv.QUOTE_ALL, 'engine': 'python'}},
        ]
        best_params = sorted([(parts['method'], parts['cols'], parts['kwargs']) for parts in params], key= lambda param_tuple: param_tuple[1], reverse=True)
        kwargs.update(best_params[0][2])

    if '.csv' in this.suffixes:
        raw = pd.read_csv(this, **kwargs)
    elif '.xlsx' in this.suffixes:
        raw = pd.read_excel(this, **kwargs)
    elif '.pkl' in this.suffixes:
        raw = pd.read_pickle(this, **kwargs)
        #return raw
    elif '.txt' in this.suffixes:
        try:
            raw = pd.read_csv(this, **kwargs) # includes '\t' if testing worked best with tabs
        except ParserError as e:
            if debug: print(f"{e}: look at file and deleted few header rows first.")
            return
        if raw.shape[1] == 1: # pandas doesn't handle \r\n two char line terminators, but seems to handle windows default if unspecified.
            raw = pd.read_csv(this, sep='\t', lineterminator='\r', **kwargs) # leaves \n in values of first column, but loads
            # lineterminator='\r\n')
            # or use codecs first to load and parse text file before dataframing...
    else:
        print(f'ERROR: this file type ({this.suffix}) is not supported')
        return
    return raw



# -*- coding: utf-8 -*-
# built ins
import logging
import os
import datetime

# package
import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from pathlib import Path

# because sklearn is a HUGE library and we're only using a single function from it,
# I've embedded the code we need in a local file. This makes methylcheck small enough to run in AWS lambda,
# but it won't use our local code copy if sklearn is installed in the environment. Just a fallback / low-env option.
try:
    from sklearn.manifold import MDS
except (ImportError, ModuleNotFoundError):
    from .sklearn_mds import *
#app
from methylcheck.progress_bar import * # tqdm, environment-specific import
from ..probes.filters import drop_nan_probes
## TODO unit-testing:
## combine_mds and ._load_data are not tested

LOGGER = logging.getLogger(__name__)

__all__ = [
    'mean_beta_plot',
    'beta_density_plot',
    'sample_plot',
    'cumulative_sum_beta_distribution', # a QC function
    'beta_mds_plot',
    'mean_beta_compare', #pre-vs-post
    'combine_mds',
]

def mean_beta_plot(df, verbose=False, save=False, silent=False):
    """Returns a plot of the average beta values for all probes in a batch of samples.

    Input (df):
        - a dataframe with probes in rows and sample_ids in columns.
        - to get this formatted import, use `methylprep.consolidate_values_for_sheet()`,
        as this will return a matrix of beta-values for a batch of samples (by default)."""
    if df.shape[0] < df.shape[1]:
        ## ensure probes in rows and samples in cols
        if verbose:
            print("Your data needed to be transposed (df = df.transpose()).")
            LOGGER.info("Your data needed to be transposed (df = df.transpose()).")
        df = df.copy().transpose() # don't overwrite the original

    data = df.copy(deep=True)
    data['mean'] = data.mean(numeric_only=True, axis=1)
    fig, ax = plt.subplots(figsize=(12, 9))
    #sns.distplot(data['mean'], hist=False, rug=False, ax=ax, axlabel='beta')
    sns.kdeplot(data['mean'], ax=ax, label='beta')
    plt.title('Mean Beta Plot')
    plt.grid()
    plt.xlim(0,1.0)
    plt.xlabel('Mean Beta')
    if save:
        plt.savefig('mean_beta.png')
    if not silent:
        plt.show()
        plt.close('all')
    else:
        plt.close(fig)



def beta_density_plot(df, verbose=False, save=False, silent=False, reduce=0.1, plot_title=None, ymax=None, return_fig=False, full_range=False, highlight_samples=None, figsize=(12,9), show_labels=None, filename='beta.png'):
    """Returns a plot of beta values for each sample in a batch of samples as a separate line.
    Y-axis values is the count (of what? intensity? normalized?).
    X-axis values are beta values (0 to 1) for a single samples

    Input (df):
        - a dataframe with probes in rows and sample_ids in columns.
        - to get this formatted import, use ``methylprep.consolidate_values_for_sheet()``, as this will return a matrix of beta-values for a batch of samples (by default).

    Returns:
        None
        (but if return_fig is True, returns the figure object instead of showing plot)

    Parameters:
        verbose:
            display extra messages
        save:
            if True, saves a copy of the plot as a png file.
        silent:
            if True, eliminates all messages (useful for automation and scripting)
        reduce:
            when working with datasets and you don't need publication quality "exact" plots,
            supply a float between 0 and 1 to sample the probe data for plotting.
            We recommend 0.1, which plots 10% of the 450k or 860k probes, and doesn't distort
            the distribution much. Values below 0.001 (860 probes out of 860k) will show some sampling distortion.
            Using 0.1 will speed up plotting 10-fold of large batches.
        ymax (None):
            If defined, upper limit of plot will not exceed this value. But it y-range can be smaller if values are less than this range.
        full_range: (False)
            if True, x-axis will be auto-scaled, instead of fixed in the 0-to-1.0 range.
        return_fig: (False)
            if True, returns figure object instead of showing plot.

        highlight_samples:
            a string or list of df col-names that, if provided, will highlight sample(s) in blue and bold in plot returned. all other samples in df will be grayed out. Useful for QC reports.
        figsize:
            tuple of width, height, with 12,9 being default if ommitted.
        show_labels: By default, sample names appear in a legend if there are <30 samples. Otherwise, ommitted. Use this to force legend on or off.

    Note:
        if the sample_ids in df.index are not unique, it will make them so for the purpose of plotting.
     """
    # ensure sample ids are unique
    if df.shape[0] > df.shape[1]:
        df = df.transpose()
        # must have samples in rows/index to reindex. you can't "recolumn".
        if len(list(df.index)) != len(list(set(df.index))):
            LOGGER.info("Your sample ids contain duplicates.")
            # rename these for this plot only
            df['ids'] = df.index.map(lambda x: x + '_' + str(int(1000000*np.random.rand())))
            df = df.set_index('ids', drop=True)
            df = df.transpose()

    if df.shape[0] < df.shape[1]:
        ## ensure probes in rows and samples in cols
        if verbose:
            LOGGER.info("Your data needed to be transposed (df = df.transpose()).")
        df = df.transpose()
    # 2nd check: incomplete probes
    if df.shape[0] < 27000:
        LOGGER.warning("Data does not appear to be full probe data")
    # 3rd check: missing probe values (common with EPIC+ and pvalues)
    pre_probe_count = df.shape[0]
    missing_probes = sum(df.isna().sum())
    if missing_probes > 0 and silent:
        df = df.copy().dropna()
    elif missing_probes > 0 and not verbose:
        LOGGER.warning(f"Your data contains {int(missing_probes/len(df.columns))} missing probe values per sample, ({missing_probes} overall). For a list per sample, use verbose=True")
        df = df.copy().dropna()
    elif missing_probes > 0 and verbose:
        LOGGER.warning(f"Your data contains {int(missing_probes/len(df.columns))} missing probe values per sample, ({missing_probes} overall).")
        LOGGER.info(df.isna().sum())
        df = df.copy().dropna()
    if df.shape[0] < 2:
        print(f"ERROR: Could not plot these samples because there are missing values for each probe in at least one sample in the batch: Started with {pre_probe_count} probes and ended with {df.shape[0]} probes shared after filtering.")
        return

    if reduce != None and reduce < 1.0:
        if not isinstance(reduce, (int, float)):
            try:
                reduce = float(reduce)
            except Exception as e:
                raise ValueError(f"{reduce} must be a floating point number between 0 and 1.0")
        # the fraction of probes in index (rows) to include in plot.
        # speeds plotting up a TON.
        # choice returns the positions as a list.
        probes = np.random.choice(df.shape[0], int(reduce*df.shape[0]))

    fig, ax = plt.subplots(figsize=figsize)

    if show_labels is None: # user-param-always overrides this.
        show_labels = True if len(df.columns) <= 30 else False
    for col in df.columns: # samples
        if col not in ('Name','ID_REF'): # probe name
            if reduce == 1.0:
                values = df[col].values
            elif reduce != None and 0 < reduce < 1.0:
                values = df[col].values[probes]
            else:
                values = df[col].values
            if len(values.shape) > 1:
                raise ValueError("Your df probaby contains duplicate sample names.")

            # updated for seaborn v0.11 (kwargs differ from v0.10; used kde_kws["lw"] instead of "linewidth")
            kde_kws = None # optional, conditional color highlighting for one or several samples in the DF, using the "highlight_samples" list passed in.
            if highlight_samples not in (None,[]):
                kde_kws={"color": "lightgray", "linewidth": 1}
                if isinstance(highlight_samples, str) and col == highlight_samples:
                    kde_kws.update({"color":"b", "linewidth":3}) #"fill": True
                if isinstance(highlight_samples, list) and col in highlight_samples:
                    kde_kws.update({"color":"b", "linewidth":3}) #"fill":True

            #kwargs = {"hist":False, "rug":False, "kde":True, "ax":ax, "axlabel":'beta'} seaborn v.10
            kwargs = {"ax": ax}
            if kde_kws:
                #kwargs["kde_kws"] = kde_kws seaborn v0.10
                kwargs.update(kde_kws) # seaborn v0.11
            if show_labels:
                kwargs["label"] = col

            if values.shape[0] < 2:
                if verbose:
                    print(f"Skipping blank sample {values}")
                continue
            #sns.distplot(values, **kwargs)
            sns.kdeplot(values, **kwargs)

    (obs_ymin, obs_ymax) = ax.get_ylim()
    #if verbose: #DEBUG for if plot is blank
    #    print(values.shape, kwargs, obs_ymin, obs_ymax)

    if ymax is not None and obs_ymax > ymax:
        ax.set_ylim(0, ymax)
    if show_labels:
        plt.legend(loc='best') # bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    #else:
    #    ax.get_legend().set_visible(False)
    #    print('suppressing legend')
    plt.title(plot_title or 'Beta Density Plot')
    plt.grid()
    if not full_range:
        plt.xlim(0,1.0)
    plt.xlabel('Beta values')
    if save:
        plt.savefig(filename, bbox_inches='tight')
    if return_fig:
        return fig
    if not silent:
        #plt.tight_layout()
        plt.show()
        plt.close('all')
    else:
        plt.clf()
        plt.cla()
        plt.close()


def sample_plot(df, **kwargs):
    """A more intuitive alias of ``beta_density_plot()``, since not all values are beta distributions.
    Note: This changes the ``beta_density_plot()` defaults to show a reduced, faster version of probe data, sampling 10% of probes present for10-fold faster processing time."""
    if 'reduce' in kwargs and kwargs.get('reduce') is None:
        reduce = None
    else:
        reduce = kwargs.get('reduce',0.1) # sets default
    kwargs['reduce'] = reduce
    beta_density_plot(df, **kwargs)


def cumulative_sum_beta_distribution(df, cutoff=0.7, verbose=False, save=False, silent=False):
    """ attempts to filter outlier samples based on the cumulative area under the curve
    exceeding a reasonable value (cutoff).

    Inputs:
        DataFrame -- wide format (probes in columns, samples in rows)
        cutoff (default 0.7)
        silent -- suppresses figure, so justs returns transformed data if False.
        if save==True: saves figure to disk.

    Returns:
        dataframe with subjects removed that exceed cutoff value."""
    # ensure probes in colums, samples in rows
    if df.shape[1] < df.shape[0]:
        df = df.copy().transpose() # don't overwrite the original
        if verbose:
            print("Your data needed to be transposed (df = df.transpose()).")
            LOGGER.info("Your data needed to be transposed (df = df.transpose()).")

    good_samples = []
    outliers = []
    if not silent:
        print("Calculating area under curve for each sample.")
    fig, ax = plt.subplots(figsize=(12, 9))
    # first, check if probes aren't consistent, to avoid a crash here
    df = drop_nan_probes(df, silent=silent, verbose=verbose)

    # if silent is True, tqdm will not show process bar.
    for subject_num, (row, subject_id) in tqdm(enumerate(zip(df.values,
                                                             df.index)), disable=silent):
        hist_vals = np.histogram(row, bins=10)[0]
        hist_vals = hist_vals / np.sum(hist_vals)
        cumulative_sum = np.cumsum(hist_vals)
        if cumulative_sum[5] < cutoff:
            good_samples.append(subject_num)
            #sns.distplot(row, hist=False, norm_hist=False)
            sns.kdeplot(row)
        else:
            outliers.append(subject_id) # drop uses ids, not numbers.

    plt.title('Cumulative Sum Beta Distribution (filtered at {0})'.format(cutoff))
    plt.grid()
    if len(df.columns) <= 30:
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    else:
        ax.legend_ = None
    if save:
        plt.savefig('cum_beta.png')
    if not silent:
        plt.show()
    plt.clf()
    plt.cla()
    plt.close()
    return df.drop(outliers, axis=0)


def beta_mds_plot(df, filter_stdev=1.5, verbose=False, save=False, silent=False, multi_params={'draw_box':True}, plot_removed=False, nafill='quick', poobah=None, palette="spectral"):
    """Performs multidimensional scaling on a dataframe of samples

Arguments
---------

    ``df``:
        dataframe of beta values for a batch of samples (rows are probes; cols are samples)
    ``filter_stdev``:
        a value (unit: standard deviations) between 0 and 3 (typically) that represents
        the fraction of samples to include, based on the standard deviation of this batch of samples.
        So using the default value of 1.5 means that all samples whose MDS-transformed beta sort_values
        are within +/- 1.5 standard deviations of the average beta are retained in the data returned.
    ``multi_params``:
        is a dict, passed into this function from a multi-compare-MDS wrapper function, containing:
        {return_plot_obj=True,
        fig=None,
        ax=None,
        draw_box=False,
        xy_lim=None,
        color_num=0,
        PSF=1.2 -- plot scale factor (margin beyond points to display)}
    ``plot_removed``:
        if True, displays a plot of samples' beta-distributions that were removed by MDS filtering.
        ignored if silent=True.
    ``nafill``: ('quick' | 'impute')
        by default, most samples will contain missing values where probes failed the signal-noise detection
        in methylprep. By default, it will use the fastest method of filling in samples from adjacent sample's probe values
        with the 'quick' method. Or, if you want it to use the average value for all samples for each probe, use 'impute', which will be much slower.
    ``poobah``: path to poobah_values.pkl file. Default is None. If supplied, this will color code dots according to
        percent of failed probes for each sample as a second dimension of QC on the plot. Does not filter or affect the output dataframe returned.

Options
--------

    ``verbose``:
        If True, provides additional messages

    ``silent``:
        - if running from command line in an automated process, you can run in `silent` mode to suppress any user interaction.
        - In this case, whatever `filter_stdev` you assign is the final value, and a file will be processed with that param.
        - Silent also suppresses plots (images) from being generated. only files are returned.

returns
-------

    Returns a filtered dataframe. If ``return_plot_obj`` is True, it returns the plot, for making overlays in ``methylize``.

requires
--------

    pandas, numpy, pyplot, sklearn.manifold.MDS

notes
-----

    this will remove probes from ALL samples in batch from consideration if any samples contain NaN (missing values) for that probe.

.. todo::

    - Reads the manifest file for the array, or at least a list of probe names to exclude/include.
    - manifest_file = pd.read_csv('/Users/nrigby/GitHub/stp-prelim-analysis/working_data/CombinedManifestEPIC.manifest.CoreColumns.csv')[['IlmnID', 'CHR']]
    - probe_names_no_sex_probes = manifest_file.loc[manifest_file['CHR'].apply(lambda x: x not in ['X', 'Y', np.nan]), 'IlmnID'].values
    - probe_names_sex_probes = manifest_file.loc[manifest_file['CHR'].apply(lambda x: x in ['X', 'Y']), 'IlmnID'].values
    df_no_sex_probes = df[probe_names_no_sex_probes]
    df_no_sex_probes.head()

    """
    # before running this, you'd typically exclude probes.
    if verbose:
        logging.basicConfig(level=logging.INFO)

    if len(df.columns) < 2:
        LOGGER.warning("beta_mds_plot requires at least 2 samples")
        return df

    if isinstance(poobah, str):
        poobah = Path(poobah)
    if isinstance(poobah, Path):
        # load the df
        if Path(poobah).is_file():
            poobah = pd.read_pickle(poobah)
        elif Path(poobah,'poobah_values.pkl').is_file():
            poobah = pd.read_pickle(Path(poobah,'poobah_values.pkl'))

    # ensure "long format": probes in rows and samples in cols. This is how methylprep returns data.
    if df.shape[1] < df.shape[0]:
        pre_df_shape = df.shape
        # MDS needs a wide matrix, with probes in columns
        df = df.transpose() # don't overwrite the original
        if verbose:
            LOGGER.info(f"Your data needed to be transposed (from {pre_df_shape} to {df.shape}) to ensure probes are in columns.")
    original_df = df.copy() # samples in index, guaranteed. transpose at end
    # require 2 or more samples for MDS

    # CHECK for missing probe values NaN -- this is common as of methylprep version 1.2.5 because pOOBah removes probes from samples by default.
    missing_probe_counts = df.isna().sum()
    total_missing_probes = round(sum([i for i in missing_probe_counts])/len(df),1) # sum / columns
    if sum([i for i in missing_probe_counts]) > 0:
        #df = df.dropna() # removes samples containing NaN probes. apparently ALL removed with poobah.
        # replace
        # random_big_ints = {i:np.random.randint(10,1000000) for i in list(df.index)}
        #random_int_df = pd.DataFrame(np.random.randint(10, 1000000, size=(df.shape[0], df.shape[1])), columns=df.columns, index=df.index)

        if nafill == 'quick': # gave best/fastest reasonable results. uses adj probe value from same sample.
            df = df.fillna(axis='index', method='bfill') # uses adjacent probe value from same sample to make MDS work.
            df = df.fillna(axis='index', method='ffill') # uses adjacent probe value from same sample to make MDS work.
            df = df.replace(np.nan, -1) # anything still NA
        elif nafill == 'impute':
            if verbose == True:
                LOGGER.info("Starting to fill in missing probes using average of that probe's value in batch of samples. This may take a while.")
            # use average value for each probe to fill in NaNs for that probe -- a much slower method.
            #probe_means = df.mean(axis=0, skipna=True) #.to_dict()
            df = df.apply(lambda x: x.fillna(x.mean()), axis=0) # per column (per probe) mean imputing.
            if sum(df.isna().sum()) > 0:
                df = df.fillna(-1)
        elif nafill == 'debug': # using adj sample probe's value instead of same sample adj probe
            df = df.fillna(axis='columns', method='bfill') # uses adjacent probe value from same sample to make MDS work.
            df = df.fillna(axis='columns', method='ffill') # uses adjacent probe value from same sample to make MDS work.
            df = df.replace(np.nan, -1) # anything still NA

        if verbose == True:
            print(f"{total_missing_probes} probe(s) [avg per sample] were missing values and removed from MDS calculations; {df.shape[1]} remaining.")
            print(f"used the 'forward-fill' method to assign values from adjacent probes so MDS can run. results are fuzzier now.")
        if silent == False:
            LOGGER.info(f"{total_missing_probes} probe(s) [avg per sample] were missing values and removed from MDS calculations; {df.shape[1]} remaining.")

    mds = MDS(n_jobs=-1, random_state=1, verbose=1)
    #n_jobs=-1 means "use all processors"
    mds_transformed = mds.fit_transform(df.values)

    # pass in df.values (a np.array) instead of a dataframe, as it processes much faster.
    # old range is used for plotting, with some margin on outside for chart readability

    # plot_scale_fator -- an empirical number to stretch the plot out and show clusters more easily.
    if 'PSF' in multi_params:
        PSF = multi_params.get('PSF')
    else:
        PSF = 2

    if df.shape[0] < 40:
        DOTSIZE = 17
    elif 40 < df.shape[0] < 60:
        DOTSIZE = 15
    elif 40 < df.shape[0] < 60:
        DOTSIZE = 13
    elif 60 < df.shape[0] < 80:
        DOTSIZE = 11
    elif 80 < df.shape[0] < 100:
        DOTSIZE = 9
    elif 100 < df.shape[0] < 300:
        DOTSIZE = 7
    else:
        DOTSIZE = 5
    old_X_range = [min(mds_transformed[:, 0]), max(mds_transformed[:, 0])]
    old_Y_range = [min(mds_transformed[:, 1]), max(mds_transformed[:, 1])]
    #old_X_range = [old_X_range[0] - PSF*old_X_range[0], old_X_range[1] + PSF*old_X_range[1]]
    #old_Y_range = [old_Y_range[0] - PSF*old_Y_range[0], old_Y_range[1] + PSF*old_Y_range[1]]
    x_std, y_std = np.std(mds_transformed,axis=0)
    x_avg, y_avg = np.mean(mds_transformed,axis=0)

    adj = filter_stdev #(1.5)

    if verbose == True:
        print("""You can now remove outliers based on their transformed beta values
 falling outside a range, defined by the sample standard deviation.""")

    while True:
        df_indexes_to_retain = []
        df_indexes_to_exclude = []
        # xy_lim is a manual override preset boundary cutoff range for MDS filtering.
        if multi_params.get('xy_lim'):
            xy_lim = multi_params['xy_lim']
            if type(xy_lim) in (list,tuple): # ((xmin, xmax), (ymin, ymax))
                minX = xy_lim[0][0]
                maxX = xy_lim[0][1]
                minY = xy_lim[1][0]
                maxY = xy_lim[1][1]
            else:
                raise ValueError ("xy_lim in multi_params must be a list or tuple")
        else:
            minX = round(x_avg - adj*x_std)
            maxX = round(x_avg + adj*x_std)
            minY = round(y_avg - adj*y_std)
            maxY = round(y_avg + adj*y_std)

        if verbose == True:
            print('Your acceptable value range: x=({0} to {1}), y=({2} to {3}).'.format(
                minX, maxX,
                minY, maxY
            ))
        # md2 are the dots that fall inside the cutoff.
        md2 = []
        for idx,row in enumerate(mds_transformed):
            if minX <= row[0] <= maxX and minY <= row[1] <= maxY:
                md2.append(row)
                df_indexes_to_retain.append(idx)
            else:
                df_indexes_to_exclude.append(idx)
            #pandas style: mds2 = mds_transformed[mds_transformed[:, 0] == class_number[:, :2]

        # this is a np array, not a df. Remove all dots that are retained from the "exluded" data set (mds_transformed))
        #mds_transformed = np.delete(mds_transformed, [df_indexes_to_retain], axis=0)
        md2 = np.array(md2)

        color_num = multi_params.get('color_num',0)
        # ADD TO EXISTING FIG... up to 24 colors (repeated 3x for 72 total) for separate QC groups
        COLORSET = dict(enumerate(['xkcd:blue', 'xkcd:green', 'xkcd:coral', 'xkcd:lightblue', 'xkcd:magenta', 'xkcd:goldenrod', 'xkcd:plum', 'xkcd:beige',
                                   'xkcd:orange', 'xkcd:orchid', 'xkcd:silver', 'xkcd:purple', 'xkcd:pink', 'xkcd:teal', 'xkcd:tomato', 'xkcd:yellow',
                                   'xkcd:olive', 'xkcd:lavender', 'xkcd:indigo', 'xkcd:black', 'xkcd:azure', 'xkcd:brown', 'xkcd:aquamarine', 'xkcd:darkblue',

                                   'xkcd:blue', 'xkcd:green', 'xkcd:coral', 'xkcd:lightblue', 'xkcd:magenta', 'xkcd:goldenrod', 'xkcd:plum', 'xkcd:beige',
                                   'xkcd:orange', 'xkcd:orchid', 'xkcd:silver', 'xkcd:purple', 'xkcd:pink', 'xkcd:teal', 'xkcd:tomato', 'xkcd:yellow',
                                   'xkcd:olive', 'xkcd:lavender', 'xkcd:indigo', 'xkcd:black', 'xkcd:azure', 'xkcd:brown', 'xkcd:aquamarine', 'xkcd:darkblue',

                                   'xkcd:blue', 'xkcd:green', 'xkcd:coral', 'xkcd:lightblue', 'xkcd:magenta', 'xkcd:goldenrod', 'xkcd:plum', 'xkcd:beige',
                                   'xkcd:orange', 'xkcd:orchid', 'xkcd:silver', 'xkcd:purple', 'xkcd:pink', 'xkcd:teal', 'xkcd:tomato', 'xkcd:yellow',
                                   'xkcd:olive', 'xkcd:lavender', 'xkcd:indigo', 'xkcd:black', 'xkcd:azure', 'xkcd:brown', 'xkcd:aquamarine', 'xkcd:darkblue']))

        # these are used with poobah as a scale of failure rate
        poobah_palettes = {
        "magma": dict(enumerate(
            [(0.15, 0.6, 0.4), (0.3, 0.25, 0.60), (0.44, 0.25, 0.5), (0.7, 0.2, 0.2), (0.973, 0.462, 0.36), (0.68, 0.60, 0.25), (0.9, 0.8, 0.2)]
        )),
        "twilight": dict(enumerate(
            [(0.5383401557517628, 0.677970811290954, 0.7711273443905772), (0.37324816143326384, 0.38140848555753937, 0.7059820097480604), (0.28669151388283165, 0.08242987384848069, 0.39331182532243375), (0.3153587000920581, 0.07942099731524319, 0.2681190407694196), (0.6456734938264543, 0.2615520316161528, 0.31248673753935263), (0.7904577684772788, 0.5982813427706246, 0.48622257801969754)]
        )),
        "viridis": dict(enumerate(
            [(0.275191, 0.194905, 0.496005), (0.212395, 0.359683, 0.55171), (0.153364, 0.497, 0.557724), (0.122312, 0.633153, 0.530398), (0.288921, 0.758394, 0.428426), (0.626579, 0.854645, 0.223353)]
        ))
        }
        sb_palette = poobah_palettes.get(palette, poobah_palettes.get("magma"))

        if multi_params.get('fig') == None:
            fig = plt.figure(figsize=(12, 9))
            plt.title('MDS Plot of betas from methylation data')
            plt.grid(linestyle='dotted') # adds fig.axes implied
        else:
            fig = multi_params.get('fig')

        if multi_params.get('ax') == None and fig.axes != []:
            if verbose:
                print('axes', multi_params.get('ax'), fig.axes, 'assigned to ax.')
            ax = fig.axes[0] # appears to get implicitly created when plot.grid() runs
        elif multi_params.get('ax') == None:
            # passing the 'ax' (fig.axes[0]) object back in will avoid the plotlib warning.
            ax = fig.add_subplot(1,1,1)
        else:
            ax = multi_params.get('ax')

        if isinstance(poobah, pd.DataFrame):
            percent_failures_hues, legend_order = _add_poobah(poobah)
            poobah_color_lookup = list(enumerate(percent_failures_hues)) # list of tuples with (0, '5 to 10') data. numbers should match md2
            color_lookup = {legend_group:[] for legend_group in legend_order}
            for N,legend_group in poobah_color_lookup:
                color_lookup[legend_group].append(N)
            ax.scatter(mds_transformed[:, 0], mds_transformed[:, 1], s=DOTSIZE, c='xkcd:ivory', edgecolor='black', linewidth=0.4)
            ax.scatter(md2[:, 0], md2[:, 1], s=DOTSIZE, c=COLORSET.get(color_num,'black'))
            color_num = 0
            for legend_group in legend_order:
                this_x = mds_transformed[color_lookup[legend_group], 0]
                this_y = mds_transformed[color_lookup[legend_group], 1]
                ax.scatter(this_x, this_y, s=DOTSIZE, color=sb_palette.get(color_num,'black'), label=legend_group)
                color_num += 1
            ax.legend(title="p-value failure (%)")
        else:
            ax.scatter(mds_transformed[:, 0], mds_transformed[:, 1], s=DOTSIZE, color='xkcd:ivory', edgecolor='black', linewidth=0.4) # EXCLUDED
            ax.scatter(md2[:, 0], md2[:, 1], s=DOTSIZE, color=COLORSET.get(color_num,'black'), edgecolor='black', linewidth=0.4) # RETAINED

        x_range_min = PSF*old_X_range[0] if PSF*old_X_range[0] < minX else PSF*minX
        x_range_max = PSF*old_X_range[1] if PSF*old_X_range[1] > maxX else PSF*maxX
        y_range_min = PSF*old_Y_range[0] if PSF*old_Y_range[0] < minY else PSF*minY
        y_range_max = PSF*old_Y_range[1] if PSF*old_Y_range[1] > maxY else PSF*maxY
        ax.set_xlim([x_range_min, x_range_max])
        ax.set_ylim([y_range_min, y_range_max])
        #print(int(x_range_min), int(x_range_max), int(y_range_min), int(y_range_max))

        if multi_params.get('draw_box') == True:
            ax.vlines([minX, maxX], minY, maxY, color=COLORSET.get(color_num,'red'), linestyle=':')
            ax.hlines([minY, maxY], minX, maxX, color=COLORSET.get(color_num,'red'), linestyle=':')

        if multi_params.get('return_plot_obj') == True:
            return fig, ax, df_indexes_to_retain

        if save:
            saved_fig = plt.gcf()
        if silent == True:
            # take the original dataframe (df) and remove samples that are outside the sample thresholds, returning a new dataframe
            df.drop(df.index[df_indexes_to_exclude], inplace=True)
            image_name = df.index.name or f'beta_mds_n={len(df.index)}'
            outfile = '{0}_s={1}_{2}.png'.format(image_name, filter_stdev, datetime.date.today())
            if save == True:
                plt.savefig(outfile)
                LOGGER.info("Saved {0}".format(outfile))
            plt.close(fig)
            # returning DataFrame in original structure: rows are probes; cols are samples.
            return df  # may need to transpose this first.
        else:
            plt.show()
            plt.close(fig)

        ########## BEGIN INTERACTIVE MODE #############
        print(f"{mds_transformed.shape[0]} original samples; {md2.shape[0]} after filtering")
        print('Your scale factor was: {0}'.format(adj))
        adj = input("Enter new scale factor, <enter> to accept and save: ")
        if adj == '':
            break
        try:
            adj = float(adj)
        except ValueError:
            print("Not a valid number. Type a number with a decimal value, or Press <enter> to quit.")
            continue

    prev_df = len(df)
    # take the original dataframe (df) and remove samples that are outside the sample thresholds, returning a new dataframe
    df_removed = df.loc[df.index[df_indexes_to_exclude]]
    df_out = df.drop(df.index[df_indexes_to_exclude]) # inplace=True will modify the original DF outside this function.
    # df_out: probes in cols and samples in index.

    if plot_removed == True and df_removed.shape[0] > 0:
        LOGGER.info(df_removed.shape)
        try:
            if df_removed.shape[1] < 50000 and df_removed.shape[0] < 50:
                sample_plot(df_removed, reduce=0.999)
            elif df_removed.shape[0] < 30:
                sample_plot(df_removed, reduce=0.999)
            else:
                sample_plot(df_removed, reduce=None)
        except:
            LOGGER.error("Could not plot removed samples.")

    if save:
        # save file. return dataframe.
        """
        fig = plt.figure(figsize=(12, 9))
        plt.title('MDS Plot of betas from methylation data')
        plt.grid() # adds fig.axes implied
        ax = fig.add_subplot(1,1,1)
        ax.scatter(md2[:, 0], md2[:, 1], s=DOTSIZE, c=COLORSET.get(color_num,'black')) # RETAINED
        ax.scatter(mds_transformed[:, 0], mds_transformed[:, 1], s=DOTSIZE, c='xkcd:ivory', edgecolor='black', linewidth=0.4,) # EXCLUDED
        ax.set_xlim(old_X_range)
        ax.set_ylim(old_Y_range)
        # UNRESOLVED BUG. (but not seen again since Sep 2019)
        # was getting 1069 samples back from this; expected 1076. discrepancy is because
        # pre_df_excl = len(df.index[df_indexes_to_exclude])
        # unique_df_excl = len(set(df.index[df_indexes_to_exclude]))
        # print(pre_df_excl, unique_df_excl)
        """

        image_name = df.index.name or f'beta_mds_n={len(df.index)}'
        outfile = '{0}_s={1}_{2}.png'.format(image_name, filter_stdev, datetime.date.today())
        saved_fig.savefig(outfile)
        plt.close(fig) # avoids displaying plot again in jupyter.
        if verbose:
            LOGGER.info("Saved {0}".format(outfile))
    # return DataFrame in original "long" format: rows are probes; cols are samples.
    df_out = original_df.drop(original_df.index[df_indexes_to_exclude])
    if df_out.shape[0] < df_out.shape[1]:
        df_out = df_out.transpose()
    return df_out


def _add_poobah(poobah):
    """ used by beta_MDS_plot """
    if poobah.isna().sum().sum() > 0:
        LOGGER.warning("Your poobah_values.pkl file contains missing values; color coding may be inaccurate.")
    percent_failures = round(100*( poobah[poobah > 0.05].count() / poobah.count() ),1)
    percent_failures = percent_failures.rename('probe_failure_(%)')
    # Series.where will replace the stuff that is False, so you have to negate it.
    percent_failures_hues = percent_failures.where(~percent_failures.between(0,5), 0)
    percent_failures_hues.where(~percent_failures_hues.between(5,10), 1, inplace=True)
    percent_failures_hues.where(~percent_failures_hues.between(10,15), 2, inplace=True)
    percent_failures_hues.where(~percent_failures_hues.between(15,20), 3, inplace=True)
    percent_failures_hues.where(~percent_failures_hues.between(20,25), 4, inplace=True)
    percent_failures_hues.where(~percent_failures_hues.between(25,30), 5, inplace=True)
    percent_failures_hues.where(~(percent_failures_hues > 30), 6, inplace=True)
    percent_failures_hues = percent_failures_hues.astype(int)
    percent_failures_hues = percent_failures_hues.replace({0:'0 to 5', 1:'5 to 10', 2:'10 to 15', 3:'15 to 20', 4:'20 to 25', 5:'25 to 30', 6:'>30'})
    legend_order = ['0 to 5','5 to 10','10 to 15','15 to 20','20 to 25','25 to 30','>30']
    #hues_palette = sb.color_palette("twilight", n_colors=7, desat=0.8) if palette is None else sb.color_palette(palette, n_colors=7, desat=0.8)
    #this = sb.scatterplot(data=qc, x="mMed", y="uMed", hue="probe_failure_(%)",
    #    palette=hues_palette, hue_order=legend_order, legend="full") # size="size"
    return percent_failures_hues, legend_order # a df with sample names in index and failure % in a column.


def mean_beta_compare(df1, df2, save=False, verbose=False, silent=False):
    """Use this function to compare two dataframes, pre-vs-post filtering and removal of outliers.
args:
    the first argument (df1) is the "pre" dataframe of samples
    the second argument (df2) is the "post" dataframe of samples

kwargs:
    verbose: additional output
    silent: suppresses figure, so no output unless save==True too."""
    if df1.shape[0] < df1.shape[1]:
        ## ensure probes in rows and samples in cols
        if verbose:
            print("Your first data set needed to be transposed (df = df.transpose()).")
            LOGGER.info("Your data needed to be transposed (df = df.transpose()).")
        df1 = df1.copy().transpose() # don't overwrite the original
    if df2.shape[0] < df2.shape[1]:
        ## ensure probes in rows and samples in cols
        if verbose:
            print("Your second data set needed to be transposed (df = df.transpose()).")
            LOGGER.info("Your data needed to be transposed (df = df.transpose()).")
        df2 = df2.copy().transpose() # don't overwrite the original

    data1 = df1.copy(deep=True)
    data1['mean'] = data1.mean(numeric_only=True, axis=1)
    data2 = df2.copy(deep=True)
    data2['mean'] = data2.mean(numeric_only=True, axis=1)

    fig, ax = plt.subplots(figsize=(12, 9))
    #line1 = sns.distplot(data1['mean'], hist=False, rug=False, ax=ax, axlabel='beta', color='xkcd:blue')
    #line2 = sns.distplot(data2['mean'], hist=False, rug=False, color='xkcd:green')
    line1 = sns.kdeplot(data1['mean'], ax=ax, label='beta', color='xkcd:blue')
    line2 = sns.kdeplot(data2['mean'], color='xkcd:green')
    plt.title('Mean Beta Plot (Compare pre (blue) vs post (green) filtering)')
    plt.grid()
    plt.xlabel('Mean Beta')
    #plt.legend([line1, line2], ['pre','post'], bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    if save:
        plt.savefig('mean_beta_compare.png')
    if not silent:
        plt.show()
    else:
        plt.close('all')


def combine_mds(*args, **kwargs):
    """To combine (or segment) datasets for multidimensional scaling analysis

--------------
how it works:
--------------

Use this function on multiple dataframes to combine datasets, or to visualize
parts of the same dataset in separate colors. It is a wrapper of `methylcheck.beta_mds_plot()` and applies
multidimensional scaling to cluster similar samples based on patterns in probe values, as well as identify
possible outlier samples (and exclude them).

    - combine datasets,
    - run MDS,
    - see how each dataset (or subset) overlaps with the others on a plot,
    - exclude outlier samples based on a composite cutoff box (the average bounds of the component data sets)
    - calculate the percent of data excluded from the group

-------
inputs:
-------

    - ``*args``:
        - pass in any number of pandas dataframes, and it will combine them into one mds plot.
        - alternatively, you may pass in a list of filepaths as strings, and it will attempt to load these files as pickles.
        but they must be pickles of pandas dataframes containing beta values or m-values

optional keyword arguments:
---------------------------

    - ``silent``: (default False)
        (automated processing mode)
        if True, suppresses most information and avoids prompting user for anything.
        silent mode processes data but doesn't show the plot.
    - ``save``: (default False)
        if True, saves the plot png to disk.
    - ``verbose``: (default False)
        if True, prints extra debug information to screen or logger.

analysis parameters:
--------------------

    - filter_stdev:
        how broadly should you retain samples? units are standard deviations, defaults to 1.5 STDEV.
        if you increase this number, fewer outlier samples will be removed.

returns:
--------

    - returns a dataframe of transformed samples

.. todo::
    - TODO: one dataframe of the retained samples, cutoff box is avg of datasets
    - TODO: each dataset's results as a transformed file
    - default: list of samples retained or excluded
    - option: a list of pyplot subplot objects
    """
    # kwargs
    save = kwargs.get('save', False)
    silent = kwargs.get('silent', False)
    verbose = kwargs.get('verbose', True)
    filter_stdev = kwargs.get('filter_stdev', 1.5)
    output_format = kwargs.get('output')
    PRINT = print if verbose else _noprint

    # check if already dataframes, or are they strings?
    list_of_dfs = list(args)
    if any([not isinstance(item, pd.DataFrame) for item in list_of_dfs]):
        if set([type(item) for item in list_of_dfs]) == {str}:
            try:
                if silent:
                    list_of_dfs = _load_data(list_of_dfs, progress_bar=False)
                elif verbose:
                    list_of_dfs = _load_data(list_of_dfs, progress_bar=True)
                else:
                    list_of_dfs = _load_data(list_of_dfs, progress_bar=False)
            except Exception as e:
                raise FileNotFoundError ("Either your files don't exist, or they are not pickled: {0}".format(e))

    # data to combine
    dfs = pd.DataFrame()
    # OPTIONAL OUTPUT: TRACK each source df's samples for plot color coding prior to merging
    # i.e. was this sample included or excluded at end?
    # maybe use a simple class here to track the data as it is being manipulated?
    sample_source = {}
    frame_transposed = {}
    # subplots: possibly useful - list of subplot objects within the figure, and their metadata.
    subplots = []
    # track x/y ranges to adjust plot area later
    xy_lims = []

    # PROCESS MDS
    fig = None
    ax = None
    for idx, df in enumerate(list_of_dfs):
        transposed = False
        if df.shape[1] > df.shape[0]: # put probes in rows
            df = df.transpose()
            transposed = True
        for sample in df.columns:
            if sample in sample_source:
                LOGGER.warning("WARNING: sample names are not unique across data sets")
            sample_source[sample] = idx
        frame_transposed[idx] = transposed

        # first, PLOT separate MDS, each with its own box.
        try:
            fig,ax,df_indexes_to_retain = beta_mds_plot(df, filter_stdev=filter_stdev, save=save, verbose=verbose, silent=silent,
                              multi_params={'return_plot_obj':True,
                                            'fig':fig,
                                            'ax':ax,
                                            'color_num':idx,
                                            'draw_box':True,
                                            'PSF':1.2,
                                            })
            subplots.append(fig)
            xy_lims.append( (fig.axes[0].get_xlim(), fig.axes[0].get_ylim()) ) # (x_range, y_range)
        except Exception as e:
            PRINT(e)

    #set max range
    x_range_min = min(item[0][0] for item in xy_lims)
    x_range_max = max(item[0][1] for item in xy_lims)
    y_range_min = min(item[1][0] for item in xy_lims)
    y_range_max = max(item[1][1] for item in xy_lims)
    fig.axes[0].set_xlim([x_range_min, x_range_max])
    fig.axes[0].set_ylim([y_range_min, y_range_max])
    #PRINT('full chart range', int(x_range_min), int(x_range_max), int(y_range_min), int(y_range_max))
    if not silent:
        plt.show()
    plt.close('all')

    # PART 2 - calculate the average MDS QC score
    #1 run the loop. calc avg MDS boundary box for group.
    #2 rerun the loop with this as the specified boundary.
    #3 calculate percent retained per dataset
    #4 overall QC score is avg percent retained per dataset.
    avg_x_range_min = int(sum(item[0][0] for item in xy_lims)/len(xy_lims))
    avg_x_range_max = int(sum(item[0][1] for item in xy_lims)/len(xy_lims))
    avg_y_range_min = int(sum(item[1][0] for item in xy_lims)/len(xy_lims))
    avg_y_range_max = int(sum(item[1][1] for item in xy_lims)/len(xy_lims))
    xy_lim = ((avg_x_range_min, avg_x_range_max), (avg_y_range_min, avg_y_range_max))
    PRINT('Average MDS window coordinate range: (x: {0}, y:{1}'.format(xy_lim[0], xy_lim[1]))
    fig = None
    ax = None
    transformed_dfs = []
    for idx, df in enumerate(list_of_dfs):
        if df.shape[1] > df.shape[0]: # put probes in rows
            df = df.transpose()
        fig,ax,df_indexes_to_retain = beta_mds_plot(df,
            filter_stdev=filter_stdev, save=save, verbose=verbose, silent=silent,
            multi_params={
                'return_plot_obj':True,
                'fig':fig,
                'ax':ax,
                'color_num':idx,
                'draw_box':True,
                'xy_lim':xy_lim,
                'PSF':1.2,
            })
        try:
            if frame_transposed[idx]:
                # this data (df) was transposed from file-orientation, so samples are in rows now; probes in columns
                df_transformed = df.iloc[df_indexes_to_retain, :] # samples in rows
                print(df.shape, df_transformed.shape)
            else:
                # iloc: first list is row index; 2nd is column index
                df_transformed = df.iloc[:, df_indexes_to_retain] # samples in columns
                print(df.shape, df_transformed.shape)
            transformed_dfs.append(df_transformed)
        except:
            import pdb;pdb.set_trace()
    fig.axes[0].set_xlim([x_range_min, x_range_max])
    fig.axes[0].set_ylim([y_range_min, y_range_max])

    # https://stackoverflow.com/questions/32213889/get-positions-of-points-in-pathcollection-created-by-scatter
    #print(len(fig.axes[0].collections)) # collections has data in every 4th list item.
    # items in the collections after the main data set are the excluded points
    all_coords = []
    retained = []
    excluded = []
    retained_sample_dfs = []
    for i in range(0, 4*len(list_of_dfs), 1):
        DD = fig.axes[0].collections[i]
        DD.set_offset_position('data')
        #print(DD.get_offsets())
        all_coords.extend(DD.get_offsets().tolist())
        #print(i, len(all_coords))
        if i % 4 == 0: # 0, 4, 8 -- this is the first data set applied to plot. (x,y plot coords)
            retained.extend( DD.get_offsets().tolist() )
            # go from plot sample x,y to idx of samples in original dfs.
        if i % 4 == 1: # 1, 5, 9, etc -- this is the second data set applied to plot.
            excluded.extend( DD.get_offsets().tolist() )
    if verbose:
        PRINT('{0} % retained overall ({1} out of {2} samples)'.format(
            round(100*len(retained) / (len(retained) + len(excluded))),
            len(retained), len(retained) + len(excluded) ))
    if not silent:
        plt.show()
    plt.close('all')

    # TODO: output_format
    return transformed_dfs


def _load_data(filepaths, progress_bar=False, tidy_it=True):
    """Loads all pickled ('.pkl') beta values dataframe files from a given folder.
older, deprecated, redundant function to methylprep.load

Output:
    A list of dataframes. It does not merge or concatenate these dataframes.
    Afterwards you can merge this this way:
    df = pd.concat(list_of_dfs_returned, axis=1, sort=False)
    provided you use the tidy_it=True flag in this function.
    Make sure the shape of the resulting dataframe has the right number of rows and columns. If not, try axis=0
Options:
    progress_bar -- If True, shows a progress bar.
    tidy_id -- If True, ensures that all files are oriented the same way in the list of datasets returned.
        All dataframes will have probes in rows and samples in columns.
    """
    dfs = []
    if progress_bar:
        _func = tqdm(filepaths, total=len(filepaths))
    else:
        _func = filepaths
    for ff in _func:
        df = pd.read_pickle(ff)
        dfs.append(df)
    if tidy_it == True:
        tidy_dfs = []
        for df in dfs:
            if df.shape[1] > df.shape[0]: # put probes in rows
                df = df.transpose()
            tidy_dfs.append(df)
        #print(f"{len(tidy_dfs)} datasets loaded.")
        return tidy_dfs
    return dfs


def _noprint(*messages):
    """ a helper function to suppress print() if not verbose mode. """
    pass



from methylcheck import mean_beta_compare
from scipy import stats
from statsmodels.distributions.empirical_distribution import ECDF
import pandas as pd
import numpy as np


def _pval_minfi(data_containers):
    # Pull M and U values
    meth = pd.DataFrame(data_containers[0]._SampleDataContainer__data_frame.index)
    unmeth = pd.DataFrame(data_containers[0]._SampleDataContainer__data_frame.index)

    for i,c in enumerate(data_containers):
        sample = data_containers[i].sample
        m = c._SampleDataContainer__data_frame.rename(columns={'meth':sample})
        u = c._SampleDataContainer__data_frame.rename(columns={'unmeth':sample})
        meth = pd.merge(left=meth,right=m[sample],left_on='IlmnID',right_on='IlmnID',)
        unmeth = pd.merge(left=unmeth,right=u[sample],left_on='IlmnID',right_on='IlmnID')

    # Create empty dataframes for red and green negative controls
    negctlsR = pd.DataFrame(data_containers[0].ctrl_red['Extended_Type'])
    negctlsG = pd.DataFrame(data_containers[0].ctrl_green['Extended_Type'])

    # Fill red and green dataframes
    for i,c in enumerate(data_containers):
        sample = str(data_containers[i].sample)
        dfR = c.ctrl_red
        dfR = dfR[dfR['Control_Type']=='NEGATIVE']
        dfR = dfR[['Extended_Type','mean_value']].rename(columns={'mean_value':sample})
        dfG = c.ctrl_green
        dfG = dfG[dfG['Control_Type']=='NEGATIVE']
        dfG = dfG[['Extended_Type','mean_value']].rename(columns={'mean_value':sample})
        negctlsR = pd.merge(left=negctlsR,right=dfR,on='Extended_Type')
        negctlsG = pd.merge(left=negctlsG,right=dfG,on='Extended_Type')

    # Reset index on dataframes
    negctlsG = negctlsG.set_index('Extended_Type')
    negctlsR = negctlsR.set_index('Extended_Type')

    # Get M and U values for IG, IR and II

    # first pull out sections of manifest (will be used to identify which probes belong to each IG, IR, II)
    manifest = data_containers[0].manifest.data_frame[['Infinium_Design_Type','Color_Channel']]
    IG = manifest[(manifest['Color_Channel']=='Grn') & (manifest['Infinium_Design_Type']=='I')]
    IR = manifest[(manifest['Color_Channel']=='Red') & (manifest['Infinium_Design_Type']=='I')]
    II = manifest[manifest['Infinium_Design_Type']=='II']

    # second merge with meth and unmeth dataframes
    IG_meth = pd.merge(left=IG,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IG_unmeth = pd.merge(left=IG,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IR_meth = pd.merge(left=IR,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IR_unmeth = pd.merge(left=IR,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    II_meth = pd.merge(left=II,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    II_unmeth = pd.merge(left=II,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')

    # Calcuate parameters
    sdG = stats.median_absolute_deviation(negctlsG)
    muG = np.median(negctlsG,axis=0)
    sdR = stats.median_absolute_deviation(negctlsR)
    muR = np.median(negctlsR,axis=0)

    # calculate p values for type 1 Red
    pIR = pd.DataFrame(index=IR_meth.index,
                   data=1 - stats.norm.cdf(IR_meth+IR_unmeth,2*muR,2*sdR),
                   columns=IR_meth.columns)

    # calculate p values for type 1 Green
    pIG = pd.DataFrame(index=IG_meth.index,
                   data=1 - stats.norm.cdf(IG_meth+IG_unmeth,2*muG,2*sdG),
                   columns=IG_meth.columns)

    # calculat4e p values for type II
    pII = pd.DataFrame(index=II_meth.index,
                  data=1-stats.norm.cdf(II_meth+II_unmeth,muR+muG,sdR+sdG),
                  columns=II_meth.columns)
    # concat and sort
    pval = pd.concat([pIR, pIG, pII])
    pval = pval.sort_values(by='IlmnID')
    return pval

def _pval_sesame(data_containers):
    # Pull M and U values
    meth = pd.DataFrame(data_containers[0]._SampleDataContainer__data_frame.index)
    unmeth = pd.DataFrame(data_containers[0]._SampleDataContainer__data_frame.index)

    for i,c in enumerate(data_containers):
        sample = data_containers[i].sample
        m = c._SampleDataContainer__data_frame.rename(columns={'meth':sample})
        u = c._SampleDataContainer__data_frame.rename(columns={'unmeth':sample})
        meth = pd.merge(left=meth,right=m[sample],left_on='IlmnID',right_on='IlmnID',)
        unmeth = pd.merge(left=unmeth,right=u[sample],left_on='IlmnID',right_on='IlmnID')

    # Separate M and U values for IG, IR and II
    # first pull out sections of manifest (will be used to identify which probes belong to each IG, IR, II)
    manifest = data_containers[0].manifest.data_frame[['Infinium_Design_Type','Color_Channel']]
    IG = manifest[(manifest['Color_Channel']=='Grn') & (manifest['Infinium_Design_Type']=='I')]
    IR = manifest[(manifest['Color_Channel']=='Red') & (manifest['Infinium_Design_Type']=='I')]
    II = manifest[manifest['Infinium_Design_Type']=='II']

    # second merge with meth and unmeth dataframes
    IG_meth = pd.merge(left=IG,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IG_unmeth = pd.merge(left=IG,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IR_meth = pd.merge(left=IR,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    IR_unmeth = pd.merge(left=IR,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    II_meth = pd.merge(left=II,right=meth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')
    II_unmeth = pd.merge(left=II,right=unmeth,on='IlmnID').drop(columns=['Infinium_Design_Type','Color_Channel']).set_index('IlmnID')

    pval = pd.DataFrame(data=manifest.index,columns=['IlmnID'])
    for i,c in enumerate(data_containers):
        funcG = ECDF(data_containers[i].oob_green['mean_value'].values)
        funcR = ECDF(data_containers[i].oob_red['mean_value'].values)
        sample = data_containers[i].sample
        pIR = pd.DataFrame(index=IR_meth.index,data=1-np.maximum(funcR(IR_meth[sample]), funcR(IR_unmeth[sample])),columns=[sample])
        pIG = pd.DataFrame(index=IG_meth.index,data=1-np.maximum(funcG(IG_meth[sample]), funcG(IG_unmeth[sample])),columns=[sample])
        pII = pd.DataFrame(index=II_meth.index,data=1-np.maximum(funcG(II_meth[sample]), funcR(II_unmeth[sample])),columns=[sample])
        p = pd.concat([pIR,pIG,pII]).reset_index()
        pval = pd.merge(pval,p)
    return pval

def detect_probes(data_containers, method='sesame', save=False, silent=True):
    """
About:
    a wrapper for the p-value probe detection methods. Tries to check inputs and rationalize them
    with methyl-suite's standard data objects.

Inputs:
    a list of sample data_containers. The dataframe must have a 'IlMnID' for the index of probe names.
    And probe names should be `cgxxxxxx` format to work with other methylcheck functions

    To create this, use:

    data_containers = methylprep.run_pipeline(data_dir,
            save_uncorrected=True,
            betas=True)
    (if there are more than 200 samples, you'll need to load the data from disk instead, as nothing will be returned)

    method:
        sesame -- use the sesame ported algorithm
        minfi -- use the minfi ported algorithm

Checks:
    data_containers must have 'meth' and 'unmeth' columns (uncorrected data)
    the values for these columns should be between 0 and 10000s
    (beta: 0 to 1; m_value: -neg to pos+ range)

Returns:
    dataframe of p-value filtered probes
    """
    if not ('unmeth' in data_containers[0]._SampleDataContainer__data_frame.columns and
            'meth' in data_containers[0]._SampleDataContainer__data_frame.columns):
            raise ValueError("Provide a list of data_containers that includes uncorrected data (with 'meth' and 'unmeth' columns, using the 'save_uncorrected' option in run_pipeline)")
    if method == 'minfi':
        pval = pval_minfi(data_containers)
    else:
        pval = pval_sesame(data_containers)

    if silent == False:
        # plot it
        # df1 and df2 are probe X sample_id matrices
        mean_beta_compare(df1, df2, save=save, verbose=False, silent=silent)
    return pval



def in_notebook():
    if hasattr(__builtins__,'__IPYTHON__'):
        return True #globals()[notebook_mod] = __import__(notebook_mod)
    else:
        return False #globals()[console_mod] = __import__(console_mod)

if in_notebook():
    from tqdm import tqdm_notebook as tqdm
else:
    from tqdm import tqdm as tqdm



from .bead_array import (
    controls_report,
    BeadArrayControlsReporter
)
from .qc_report import (
    run_pipeline,
    ReportPDF
)

__all__ = [
    'controls_report',
    'BeadArrayControlsReporter',
    'run_pipeline',
    'ReportPDF',
]



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from collections import Counter
import methylcheck

# dummy function for pytest to test assign()
def get_input(text):
    return input(text)

def assign(df, two_pass=False):
    """Manually and interactively assign each sample to a group, based on beta-value distribution shape.

how:
    sorts samples by the position of their peak intensity in beta dist, then lets the human assign a number to
    each.
    assumes that peaks that differ by > 0.1 of the 0 to 1 range are different clusters. fills in expected cluster #.

options:
    if two_pass==True: it lets user go through every cluster a second time and split larger clusters further.
    sometimes a cluster isn't obviously two groups until it is separated from the larger dataset.
    """
    # calculate the highest bin of the sample beta histogram for each sample
    def calc_max_bin_of_hist(nparray):
        hist_vals, bin_edges = np.histogram(nparray, bins=16)
        # want the bin_edge corresponding to the max hist_val
        bin_of_max_val = np.where(hist_vals == np.amax(hist_vals))
        return bin_of_max_val[0][0] # it could occur in two bins; if so, return the first bin only
    colors = {-1:"xkcd:scarlet", 2:'xkcd:brick red', 1:"xkcd:blue", 2:"xkcd:green", 3:'xkcd:gold', 4:'xkcd:purple', 5:'xkcd:grey blue',
        6:'xkcd:brown', 7:'xkcd:bubblegum', 8:'xkcd:tangerine', 9:'xkcd:poop', 10:"xkcd:baby blue", 11:"xkcd:violet", 12:"xkcd:apple green",
        13:"xkcd:neon pink", 14:"xkcd:deep green", 15:"xkcd:blush", 16:"xkcd:lemon"}
    def sample_plot_color_clusters(df, colormap=None, verbose=False, save=False, silent=False, reduce=0.01, ymax=5, figsize=(14,8)):
        """ colormap: dict of sample_ids --> group number (some int). Sample_ids must match df columns"""
        #qualitative_colors = sns.color_palette("Set3", 10)
        #sns.set_color_codes()
        # ensure sample ids are unique
        if df.shape[0] > df.shape[1]:
            df = df.transpose()
            # must have samples in rows/index to reindex. you can't "recolumn".
            if list(df.index) != list(set(df.index)):
                LOGGER.info("Your sample ids contain duplicates.")
                # rename these for this plot only
                df['ids'] = df.index.map(lambda x: x + '_' + str(int(1000000*np.random.rand())))
                df = df.set_index('ids', drop=True)
                df = df.transpose()

        if df.shape[0] < df.shape[1]:
            ## ensure probes in rows and samples in cols
            df = df.transpose()
        if reduce != None and reduce < 1.0:
            if not isinstance(reduce, (int, float)):
                try:
                    reduce = float(reduce)
                except Exception as e:
                    raise ValueError(f"{reduce} must be a floating point number between 0 and 1.0")
            probes = np.random.choice(df.shape[0], int(reduce*df.shape[0]))
        fig, ax = plt.subplots(figsize=figsize) #was (12, 9)
        for idx, sample_id in enumerate(df.columns): # samples
            if sample_id != 'Name': # probe name
                if reduce:
                    values = df[sample_id].values[probes]
                else:
                    values = df[sample_id].values
                if len(values.shape) > 1:
                    raise ValueError("Your df probaby contains duplicate sample names.")

                this_color = colors[-1] if colormap is None else colors.get(colormap.get(sample_id,-1),'k')
                if idx+1 == len(df.columns):
                    kde_kws = {'linestyle':'--', 'linewidth':2}
                else:
                    kde_kws = {'linewidth':1}

                #sns.distplot(
                #    values, hist=False, rug=False, kde=True,
                #    ax=ax, axlabel='beta', color=this_color, kde_kws=kde_kws)
                sns.kdeplot(values, ax=ax, label='beta', color=this_color, **kde_kws)
        (data_ymin,data_ymax) = ax.get_ylim()
        if data_ymax > ymax:
            ax.set_ylim(0,ymax)
        plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
        plt.show()

    # orient so samples in columns
    if len(df.columns) > df.shape[0]:
        df = df.transpose() # now more rows(probes) than cols(samples)

    def color_guide(user_defined_groups):
        color_names = {k: v.replace('xkcd:','') for k,v in colors.items()}
        color_count = Counter(user_defined_groups.values())
        guide = []
        for group,count in color_count.most_common():
            guide.append(f"{group} {color_names[group]}")
        return ' | '.join(guide)

    sample_predicted_groups = {} # sample_name --> bin where peak occurs | as a 0-10 number for clustering
    for sample_id in df.columns: # columns
        #betas[sample] = vectorized_func(meth, unmeth)
        sample_predicted_groups[sample_id] = calc_max_bin_of_hist(df[sample_id].values)
    print('Predicted sample groups, based on beta-peak position')
    prediction_note = " | ".join([f"{k}: {v}" for k,v in Counter(sample_predicted_groups.values()).most_common()])
    print(prediction_note)
    # sort by cluster frequency
    sample_predicted_groups_counts = Counter(sample_predicted_groups.values())
    sample_predicted_groups = dict(sorted(sample_predicted_groups.items(), key=lambda kv: sample_predicted_groups_counts[kv[1]], reverse=True))

    # go through each sample, plotting it, and letting user assign to one of the groups
    user_defined_groups = {}
    colormap = {}
    past_samples = []
    for sample_id,predicted_group in sample_predicted_groups.items():
        past_samples.append(sample_id)
        sub_sample_df = pd.DataFrame(data = df[past_samples], index = df.index).transpose()
        sample_plot_color_clusters(sub_sample_df, colormap=colormap, reduce=0.01)
        print(color_guide(user_defined_groups))
        while True:
            g = get_input(f"Group ({predicted_group}): ")
            if g in (""," "):
                g = predicted_group
            try:
                g = int(g)
            except:
                print('Only use numbers between -1 and 16!')
                g = get_input(f"Group ({predicted_group}): ")
            if g not in colors:
                print('Only use numbers between -1 and 16!')
                g = get_input(f"Group ({predicted_group}): ")
            else:
                break
        user_defined_groups[sample_id] = g
        if g == -1:
            g = 10
        colormap[sample_id] = g

    # finally, produce a table and plot of each cluster:
    clusters = list(set(user_defined_groups.values()))
    cluster_samples = {}
    for cluster in clusters:
        samples = [sample_id for sample_id in df.columns if user_defined_groups.get(sample_id) == cluster]
        sub_sample_df = pd.DataFrame(data = df[samples], index = df.index)
        print(f"Group ({cluster}) {len(samples)} samples: {samples}")
        methylcheck.sample_plot(sub_sample_df)
        if two_pass:
            yesno = get_input(f"Rework Group ({cluster}) (y/n)?")
            if yesno in ('y','Y','yes'):
                past_samples = []
                for sample_id in samples:
                    past_samples.append(sample_id)
                    sub_sample_df = pd.DataFrame(data = df[past_samples], index = df.index).transpose()
                    sample_plot_color_clusters(sub_sample_df, colormap=colormap, reduce=0.01)
                    print(color_guide(user_defined_groups))
                    while True:
                        g = get_input(f"Group ({cluster}): ")
                        if g in (""," "):
                            g = cluster
                        try:
                            g = int(g)
                        except:
                            print('Only use numbers between -1 and 16!')
                            g = get_input(f"Group ({cluster}): ")
                        if g not in colors:
                            print('Only use numbers between -1 and 16!')
                            g = get_input(f"Group ({cluster}): ")
                        else:
                            break
                    user_defined_groups[sample_id] = g
                    if g == -1:
                        g = 10
                samples = [sample_id for sample_id in df.columns if user_defined_groups.get(sample_id) == cluster]
                sub_sample_df = pd.DataFrame(data = df[samples], index = df.index)
                print(f"Group ({cluster}) {len(samples)} samples: {samples} REVISED")
                methylcheck.sample_plot(sub_sample_df)
        cluster_samples[cluster] = samples
    return user_defined_groups

def plot_assigned_groups(df, user_defined_groups):
    """ takes the 'sample: group' dictionary and plots each sub-set.
    returns a lookup dict of each cluster --> [list of samples]."""
    # finally, produce a table and plot of each cluster:
    clusters = list(set(user_defined_groups.values()))
    cluster_samples = {}
    for cluster in clusters:
        samples = [sample_id for sample_id in df.columns if user_defined_groups.get(sample_id) == cluster]
        sub_sample_df = pd.DataFrame(data = df[samples], index = df.index)
        plot_title = f"Group ({cluster}) {len(samples)} samples"
        methylcheck.beta_density_plot(sub_sample_df, plot_title=plot_title)
        cluster_samples[cluster] = samples
    return cluster_samples



import pandas as pd
from pathlib import Path
from io import StringIO
import re
import sys
import logging
import math
LOGGER = logging.getLogger(__name__)
# app
import methylcheck

__all__ = ['run_pipeline', 'ReportPDF']

def run_pipeline(df, **kwargs):
    """Run a variety of probe and sample filters in tandem, then plot results

by specifying all of your options at once, instead of running every part of methylcheck
in piacemeal fashion.

this is analogous to using the methylcheck CLI, but for notebooks/scripts

required:
    df: (required)
        - data as a DataFrame of beta values
        - sample names in columns and probes in rows

parameters:
    verbose: (True/False)
        default: False -- shows extra info about processing if True
    silent: (True/False)
        default: False -- suppresses all warnings/info

    exclude_sex:
        filters out probes on sex-chromosomes
    exclude_control:
        filters out illumina control probes
    exclude_all:
        filters out the most probes (sex-linked, control, and all sketchy-listed probes from papers)
    exclude: (list of strings, shorthand references to papers with sketchy probes to exclude)

        If the array is 450K the publications may include:
            ``'Chen2013'
            'Price2013'
            'Zhou2016'
            'Naeem2014'
            'DacaRoszak2015'``
        If the array is EPIC the publications may include:
            ``'Zhou2016'
            'McCartney2016'``
        or these reasons:
            ``'Polymorphism'
            'CrossHybridization'
            'BaseColorChange'
            'RepeatSequenceElements'``
        or use ``'exclude_all'``:
            to do maximum filtering, including all of these papers' lists.

    plot: (list of strings)
        ['mean_beta_plot', 'beta_density_plot', 'cumulative_sum_beta_distribution', 'beta_mds_plot', 'all']
        if 'all', then all of these plots will be generated. if omitted, no plots are created.
    save_plots: (True|False)
        default: False
    export (True|False):
        default: False -- will export the filtered df as a pkl file if True

note:
    this pipeline cannot also apply the array-level methylcheck.run_qc() function
    because that relies on additional probe information that may not be present. Everything
    in this pipeline applies to a dataframe of beta or m-values for a set of samples.

returns:
    a filtered dataframe object
    """
    # verify kwargs are expected strings
    param_list ={
    'exclude': ['Chen2013',
                'Price2013',
                'Zhou2016',
                'Naeem2014'
                'DacaRoszak2015',
                'Zhou2016',
                'McCartney2016',
                'Polymorphism',
                'CrossHybridization',
                'BaseColorChange',
                'RepeatSequenceElements'],
    'verbose': [True, False],
    'silent': [True, False],
    'exclude_sex': [True, False],
    'exclude_control': [True, False],
    'exclude_all': [True, False],
    'plot': ['mean_beta_plot', 'beta_density_plot', 'cumulative_sum_beta_distribution', 'beta_mds_plot', 'all'],
    'save_plots': [True, False],
    'export': [True, False],
    'run_qc': [True, False],
    'on_lambda': [True, False],
    }
    for k,v in kwargs.items():
        if k not in param_list:
            raise KeyError(f"{k} not an option among {list(param_list.keys())}")
        possible_values = param_list[k]
        if type(v) is list:
            for item in v:
                if item not in possible_values:
                    raise ValueError(f"{item} not a valid option for {k}: {possible_values}")
        else:
            if v not in possible_values:
                raise ValueError(f"{item} not a valid option for {k}: {possible_values}")
    # now we know all inputs are among the allowed/expected params.

    # set detail level of messages
    if kwargs.get('verbose') == True:
        logging.basicConfig(level=logging.INFO)
    else:
        kwargs['verbose'] = False
        logging.basicConfig(level=logging.WARNING)
    if kwargs.get('silent') == True:
        logging.basicConfig(level=logging.ERROR)
    else:
        kwargs['silent'] = False

    # assign some default pipeline params, if not specified by user
    default_params = {
        'verbose': False,
        'silent': False,
        'exclude_all': True,
        'plot': 'all',
        'save_plots': False,
        'export': False,
        'on_lambda': False}
    for param,setting in default_params.items():
        if kwargs.get(param) is None:
            kwargs[param] = setting

    # determine array type
    array_type = methylcheck.detect_array(df, on_lambda=kwargs.get('on_lambda'))

    # apply some filters
    if kwargs.get('exclude_all'):
        try:
            df = methylcheck.exclude_sex_control_probes(df, array_type, verbose=kwargs.get('verbose'))
            sketchy_probe_list = methylcheck.list_problem_probes(array_type)
            df = methylcheck.exclude_probes(df, sketchy_probe_list)
        except ValueError:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            if "probe list" in str(exc_value) or "probe exclusion lists" in str(exc_value):
                # keep df the same instead of filtering.
                LOGGER.info(exc_value)
            else:
                raise Exception(exc_value)
    elif kwargs.get('exclude'):
        # could be a list or a string
        if type(kwargs['exclude']) is not list:
            kwargs['exclude'] = [kwargs['exclude']]
        try:
            sketchy_probe_list = methylcheck.list_problem_probes(array_type, criteria=kwargs['exclude'])
            df = methylcheck.exclude_probes(df, sketchy_probe_list)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            if "probe list" in str(exc_value) or "probe exclusion lists" in str(exc_value):
                LOGGER.info(exc_value)
            else:
                raise Exception(exc_value)

    # apply some more filters
    if kwargs.get('exclude_sex') and kwargs.get('exclude_control'):
        df = methylcheck.exclude_sex_control_probes(df, array_type, verbose=kwargs.get('verbose'))
    elif kwargs.get('exclude_sex'):
        df = methylcheck.exclude_sex_control_probes(df, array_type, no_sex=True, no_control=False, verbose=kwargs.get('verbose'))
    elif kwargs.get('exclude_control'):
        df = methylcheck.exclude_sex_control_probes(df, array_type, no_sex=False, no_control=True, verbose=kwargs.get('verbose'))

    # plot stuff
    # run calculations and plots
    if 'all' in kwargs.get('plot'):
        methylcheck.mean_beta_plot(df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        methylcheck.beta_density_plot(df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        wide_df = df.copy().transpose()
        methylcheck.cumulative_sum_beta_distribution(wide_df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        pre_mds = wide_df.shape[0]
        df = methylcheck.beta_mds_plot(wide_df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        LOGGER.info(f"MDS filtering dropped {pre_mds - df.shape[1]} samples from dataframe.")
    else:
        if 'mean_beta_plot' in args.plot:
            methylcheck.mean_beta_plot(df, verbose=kwargs['verbose'], save=kwargs.get('save_plots'), silent=kwargs['silent'])
        if 'beta_density_plot' in args.plot:
            methylcheck.beta_density_plot(df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        if 'cumulative_sum_beta_distribution' in args.plot:
            wide_df = df.copy().transpose()
            methylcheck.cumulative_sum_beta_distribution(wide_df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
        if 'beta_mds_plot' in args.plot:
            wide_df = df.copy().transpose()
            pre_mds = wide_df.shape[0]
            df = methylcheck.beta_mds_plot(wide_df, verbose=kwargs['verbose'], save=kwargs['save_plots'], silent=kwargs['silent'])
            LOGGER.info(f"MDS filtering dropped {pre_mds - df.shape[1]} samples from dataframe.")
            # also has filter_stdev params to pass in.
    if kwargs.get('export') == True:
        outfile = 'beta_values_filtered.pkl' # in long format for faster loading
        df.to_pickle(outfile)
        LOGGER.info(f'Saved {outfile}')
    return df


def detection_poobah(poobah_df, pval_cutoff=0.05):
    """Given a dataframe of p-values with sample_ids in columns and probes in index,
    calculates the percent PASSING probes per sample. Part of QC report.

    v0.7.2: if 'gct_score' is enabled, add column for GCT % per sample.
    v0.7.2: changed from percent failing to percent passing"""
    out = {}
    for sample_id in poobah_df.columns:
        total = poobah_df.shape[0]
        passing = poobah_df[sample_id][poobah_df[sample_id] <= pval_cutoff].count()
        percent_passing = round(100*passing/total,1)
        out[sample_id] = percent_passing
    return out


class ReportPDF:
    """ReportPDF allows you to build custom QC reports.

-------
To use:
-------

- First, initialize the report and pass in kwargs, like ``myReport = ReportPDF(**kwargs)``
- Next, run ```myReport.run_qc()`` to fill it in.
- Third, you must run ``myReport.pdf.close()`` after ``run_qc()`` to save the file to disk.
- You can supply kwargs to specify which QC plots to include
  - and supply a list of chart names to control the order of objects in the report.
  - if you pass in 'order' in kwargs, any page you omit in the order will be omitted from the final report.
  - You may pass in a custom table to override one of the built-in pages.
- include 'path' with the path to your processed pickle files.
- include an optional 'outpath' for where to save the pdf report.

-------
kwargs:
-------

- processing params
  - filename
  - poobah_min_percent (e.g. at least 80% of probes must pass for sample to pass)
  - pval_cutoff (e.g. set alpha at 0.05)
  - outpath
  - path
  - runme: Default is not to actually generate all the parts of PDF with report.run_qc() then report.pdf.close(), but setting this to True will do everything at once.
- front page text
  - title
  - author
  - subject
  - keywords
- if 'debug=True' is in kwargs,
  - then it will return a report without any parts that failed.
- tests
  - poobah: includes a table with each sample and percent of probes that passed p-value signal detection
  - gct: includes GCT scores (bisulfite conversion completeness) in poobah table
  - mds: performs multidimensional scaling to identify and report on sample outliers
- plots
  - beta_density_plot
  - M_vs_U (default False)
  - M_vs_U_compare (default False) -- shows the effect of all processing steps vs raw intensity
  - qc_signal_intensity
  - controls (A batter of ported Genome Studio plots)
  - probe_types

--------------
custom tables:
--------------

pass in arbitrary data using kwarg ``custom_tables`` as list of dictionaries with this structure:

```python
custom_tables=[
    {
    'title': "some title, optional", # NOTE: chart titles must be unique!
    'col_names': ["<list of strings>"],
    'row_names': ["<list of strings, optional>"],
    'data': ["<list of lists, with order matching col_names>"],
    'order_after': "string name of the plot this should come after. It cannot appear first in list.",
    'font_size': <can be None, int, 'auto' (shrink to page), or 'truncate' (chop of long values to make fit)>
    },
    {"<...second table here...>"}
]
```

------------------------
Pre-processing pipeline:
------------------------

    Probe-level (w/explanations of suggested exclusions)
        - Links to recommended probe exclusion lists/files/papers
        - Background subtraction and normalization ('noob')
        - Detection p-value ('neg' vs 'oob')
        - Dye-bias correction (from SeSAMe)
    Sample-level (w/explanations of suggested exclusions)
        Detection p-value (% failed probes)
        - custom detection (% failed, of those in a user-defined-list supplied to function)
        MDS
        Suggested for customer to do on their own
        - Sex check
        - Age check
        - SNP check
    """
    # larger font
    # based on 16pt with 0.1 (10% of page) margins around it: use 80, 26, 16
    # MAXWIDTH = 80 # based on 16pt with 0.1 (10% of page) margins around it
    # MAXLINES = 26
    # FONTSIZE = 16
    # ORIGIN = (0.1, 0.1) # achored on page in lower left
    #
    # normal font -- for 12pt font: use 100 x 44 lines
    MAXWIDTH = 100
    MAXLINES = 44
    FONTSIZE = 12
    ORIGIN = (0.1, 0.05) # achored on page in lower left

    def __init__(self, **kwargs):
        # https://stackoverflow.com/questions/8187082/how-can-you-set-class-attributes-from-variable-arguments-kwargs-in-python
        self.__dict__.update(kwargs)
        self.debug = True if self.__dict__.get('debug') == True else False
        self.__dict__['poobah_min_percent'] = self.__dict__.get('poobah_min_percent', 80)
        self.__dict__['pval_cutoff'] = self.__dict__.get('pval_cutoff', 0.05)
        self.errors = self.open_error_buffer()

        #SHELVING the PDF_PART, since it worked but not neeeded right now
        from matplotlib.backends.backend_pdf import PdfPages
        self.outfile = Path(self.__dict__.get('outpath','.'), self.__dict__.get('filename', 'multipage_pdf.pdf'))
        self.pdf = PdfPages(self.outfile)
        import matplotlib.pyplot as plt
        self.plt = plt
        import textwrap
        self.textwrap = textwrap
        import datetime
        self.today = str(datetime.date.today())
        self.on_lambda = self.__dict__.get('on_lambda', False)
        d = self.pdf.infodict()
        if any(kwarg in ('title','author','subject','keywords') for kwarg in kwargs):
            # set the file's metadata via the PdfPages object:
            d['Title'] = self.__dict__.get('title','')
            d['Author'] = self.__dict__.get('author','')
            d['Subject'] = self.__dict__.get('subject','')
            d['Keywords'] = self.__dict__.get('keywords','')
        d['CreationDate'] = datetime.datetime.today()
        d['ModDate'] = datetime.datetime.today()

        if 'order' in self.__dict__:
            self.order = self.__dict__['order']
        else:
            self.order = ['beta_density_plot', 'detection_poobah', 'predict_sex', 'mds', 'auto_qc',
            'qc_signal_intensity', 'M_vs_U_compare', 'M_vs_U', 'controls', 'probe_types']

        self.tests = {
            'detection_poobah': self.__dict__.get('poobah',True),
            'mds': self.__dict__.get('mds',True), # part of detection_poobah table
            'auto_qc': self.__dict__.get('auto_qc',True), # algorithmic failure prediction, NOT IMPLEMENTED YET #
            # v0.7.3 adds bis-conversion completeness, the GCT score table --- https://rdrr.io/bioc/sesame/src/R/sesame.R
            'gct_score': self.__dict__.get('gct',True), # part of detection_poobah table
        }
        self.plots = {
            'beta_density_plot': self.__dict__.get('beta_density_plot',True),
            'qc_signal_intensity': self.__dict__.get('qc_signal_intensity',True),
            'M_vs_U_compare': self.__dict__.get('M_vs_U_compare',False),
            'M_vs_U': self.__dict__.get('M_vs_U',False),
            'predict_sex': self.__dict__.get('predict_sex',False),
            'controls': self.__dict__.get('controls',True), # genome studio plots
            'probe_types': self.__dict__.get('probe_types',True),
        }

        self.custom = {}
        if 'custom_tables' in self.__dict__:
            self.parse_custom_tables(self.__dict__['custom_tables'])
            LOGGER.info(f"Found custom_tables and inserted into order: {self.order}")
            #LOGGER.info(self.custom)

        if self.__dict__.get('runme') == True:
            self.run_qc()
            self.pdf.close()

    def parse_custom_tables(self, tables):
        """tables is a list of {
        'title': "some title, optional",
        'col_names': [list of strings],
        'row_names': [list of strings, optional],
        'data': [list of lists, with order matching col_names],
        'order_after': string name of the plot this should come after. It cannot appear first in list.
        }"""
        for table in tables:
            required_attributes = {'title', 'data', 'order_after', 'col_names'}
            for i in required_attributes:
                if i not in table:
                    raise KeyError("Your custom table must contain these keys: {required_attributes} (row_names is optional)")
            #1 place order -- refer to this later in self.custom
            try:
                _index = self.order.index(table['order_after']) + 1
                self.order.insert(_index, table['title'])
                self.custom[table['title']] = table
            except ValueError:
                raise ValueError(f"Your custom table's 'order_after' label is not in this list of chart objects, so could not be ordered: {self.order}")
            if not isinstance(table['col_names'], list):
                raise TypeError(f"Your custom table 'col_names' must be a list.")
            if table.get('row_names') and not isinstance(table['row_names'], list):
                raise TypeError(f"Your custom table 'row_names' must be a list.")
            if not isinstance(table['data'], list):
                raise TypeError(f"Your custom table 'data' must be a list of lists, matching the number of columns listed in col_names.")
                if not (isinstance(table['data'][0],list) and len(table['data'][0]) == len(table['col_names'])):
                    raise TypeError(f"Your custom table 'data' must be a list of lists, matching the number of columns listed in col_names.")
            # passes! later, the table will be looked-up by title name and fed into chart in order.

    def open_error_buffer(self):
        """ preparing a stream of log messages to add to report appendix. """
        appendix = logging.getLogger()
        appendix.setLevel(logging.INFO)
        errors = StringIO()
        error_handler = logging.StreamHandler(errors)
        # %(module)s.%(funcName)s() ... %Y-%m-%d
        formatter = logging.Formatter('^^^%(asctime)s - %(levelname)s - %(message)s', "%H:%M:%S")
        error_handler.setFormatter(formatter)
        appendix.addHandler(error_handler)
        return errors

    def run_qc(self):
        # load all the data from self.path; make S3-compatible too.
        path = self.__dict__.get('path','')
        if self.tests['detection_poobah'] is True and 'detection_poobah' in self.order:
            try:
                poobah_df = pd.read_pickle(Path(path,'poobah_values.pkl')) # off by default, process with --export_poobah
            except FileNotFoundError:
                raise FileNotFoundError(f"Could not load pickle file: 'poobah_values.pkl'. Add 'poobah=False' to inputs to turn this off.")
        if self.tests['gct_score'] is True and self.tests['detection_poobah'] is False:
            LOGGER.warning(f"Cannot include GCT score unless poobah table is also enabled.")
        try:
            beta_df = pd.read_pickle(Path(path,'beta_values.pkl'))
            if Path(path,'noob_meth_values.pkl').exists() and Path(path,'noob_unmeth_values.pkl').exists():
                meth_df = pd.read_pickle(Path(path,'noob_meth_values.pkl'))
                unmeth_df = pd.read_pickle(Path(path,'noob_unmeth_values.pkl'))
            else:
                meth_df = pd.read_pickle(Path(path,'meth_values.pkl'))
                unmeth_df = pd.read_pickle(Path(path,'unmeth_values.pkl'))
            control_dict_of_dfs = pd.read_pickle(Path(path,'control_probes.pkl'))
            # for M_vs_U_compare only, need both
            if self.__dict__.get('M_vs_U_compare') == True:
                if not (Path(path,'noob_meth_values.pkl').exists() and
                    Path(path,'noob_unmeth_values.pkl').exists() and
                    Path(path,'meth_values.pkl').exists() and
                    Path(path,'unmeth_values.pkl').exists()):
                    LOGGER.error("Not all meth/unmeth pickle files found for M_vs_U compare plot. Skipping this plot.")
                    self.__dict__.pop('M_vs_U_compare')
            LOGGER.info("Data loaded")
        except FileNotFoundError:
            raise FileNotFoundError("Could not load pickle files")

        '''
        self.functions = { # NOT USED --  needed to provide different conditions for each function below. #
            'detection_poobah': detection_poobah,
            'mds': methylcheck.beta_mds_plot,
            'auto_qc': None,
            'beta_density_plot': methylcheck.beta_density_plot,
            'M_vs_U': methylcheck.plot_M_vs_U,
            'controls': methylcheck.plot_controls,
            'qc_signal_intensity': methylcheck.qc_signal_intensity,
        }
        '''
        if 'mds' in self.tests and len(beta_df.columns) > 1:
            # some things must be calculated ahead of time, because used twice
            beta_mds_fig, ax, df_indexes_to_retain = methylcheck.beta_mds_plot(beta_df, silent=True, multi_params={'return_plot_obj':True, 'draw_box':True})
            mds_passing = [sample_id for idx,sample_id in enumerate(beta_df.columns) if idx in df_indexes_to_retain]
            print(mds_passing)
            include_mds = True
        else:
            beta_mds_fig = None
            mds_passing = []
            include_mds = False

        for part in self.order:
            if part in self.tests and self.tests[part] == True:
                try:
                    if part == 'detection_poobah':
                        min_allowed = self.__dict__.get('poobah_min_percent')
                        sample_percent_failed_probes_dict = detection_poobah(poobah_df, pval_cutoff=self.__dict__['pval_cutoff'])
                        if self.tests['gct_score']:
                            sample_gct_percent_dict = methylcheck.bis_conversion_control(meth_df)
                        else:
                            sample_gct_percent_dict = {}
                        sample_poobah_failures = {sample_id: ("pass" if percent > min_allowed else "fail") for sample_id,percent in sample_percent_failed_probes_dict.items()}
                        #LOGGER.info(f"Poobah: {len([k for k in sample_poobah_failures.values() if k =='fail'])} failure(s) out of {len(sample_poobah_failures)} samples.")

                        list_of_lists = []
                        col_names=['Sample_ID', 'Percent', 'Pass/Fail']
                        if self.tests['gct_score']:
                            col_names=['Sample_ID', 'GCT score', 'Percent', 'Pass/Fail']
                        if include_mds:
                            col_names=['Sample_ID', 'MDS', 'Percent', 'Pass/Fail']
                        if self.tests['gct_score'] and include_mds:
                            col_names=['Sample_ID', 'GCT score', 'MDS', 'Poobah (%)', 'Poobah Pass/Fail']

                        for sample_id,percent in sample_percent_failed_probes_dict.items():
                            if self.tests['gct_score'] and sample_id in sample_gct_percent_dict:
                                gct_score = sample_gct_percent_dict[sample_id]
                            else:
                                gct_score = None
                            mds_pass_fail = 'Pass' if sample_id in mds_passing else 'Fail'

                            if gct_score is not None and include_mds:
                                row = [sample_id, gct_score, mds_pass_fail, percent, sample_poobah_failures[sample_id]]
                            elif gct_score is not None and not include_mds:
                                row = [sample_id, gct_score, percent, sample_poobah_failures[sample_id]]
                            elif gct_score is None and include_mds:
                                row = [sample_id, mds_pass_fail, percent, sample_poobah_failures[sample_id]]
                            elif gct_score is None and not include_mds:
                                row = [sample_id, percent, sample_poobah_failures[sample_id]]

                            list_of_lists.append(row)
                        self.to_table(list_of_lists, col_names=col_names,
                            row_names=None, add_title='Sample probe quality testing')

                    if part == 'mds' and len(beta_df.columns) > 1:
                        LOGGER.info("Beta MDS Plot")
                        # ax and df_to_retain are not used, but could go into a qc chart
                        self.pdf.savefig(beta_mds_fig)
                        self.plt.close()
                        #pretty_table = Table(titles=['Sample_ID', 'MDS Pass/Fail'])

                except Exception as e:
                    if self.debug:
                        LOGGER.error(f"Could not process {part}; {e}")
                        continue
                    else:
                        raise Exception(f"Could not process {part}; {e}")

            elif part in self.plots and self.plots[part] == True:
                try:
                    if part == 'beta_density_plot':
                        LOGGER.info("Beta Density Plot")
                        fig = methylcheck.beta_density_plot(beta_df, save=False, silent=True, verbose=False, reduce=0.1, plot_title=None, ymax=None, return_fig=True)
                        self.pdf.savefig(fig)
                        self.plt.close()
                    elif part == 'M_vs_U':
                        LOGGER.info(f"M_vs_U plot")
                        fig = methylcheck.plot_M_vs_U(meth=meth_df, unmeth=unmeth_df, noob=True, silent=True, verbose=False, plot=True, compare=False, return_fig=True, poobah=poobah_df)
                        self.pdf.savefig(fig)
                        self.plt.close()
                        # if not plotting, it will return dict with meth median and unmeth median.
                    elif part == 'M_vs_U_compare':
                        LOGGER.info(f"M_vs_U_compare plot")
                        fig = methylcheck.plot_M_vs_U(data_containers_or_path=path, compare=True, silent=False, verbose=True, plot=True, return_fig=True)
                        self.pdf.savefig(fig)
                        self.plt.close()
                        # if not plotting, it will return dict with meth median and unmeth median.
                    elif part == 'qc_signal_intensity':
                        LOGGER.info(f"QC signal intensity plot")
                        fig = methylcheck.qc_signal_intensity(meth=meth_df, unmeth=unmeth_df, silent=True, return_fig=True, poobah=poobah_df)
                        self.pdf.savefig(fig)
                        self.plt.close()
                    elif part == 'controls':
                        LOGGER.info(f"Control probes")
                        list_of_figs = methylcheck.plot_controls(control_dict_of_dfs, 'all', return_fig=True)
                        for fig in list_of_figs:
                            self.pdf.savefig(figure=fig, bbox_inches='tight')
                        self.plt.close('all')
                    elif part == 'probe_types':
                        LOGGER.info(f"Betas by probe type")
                        list_of_figs = methylcheck.plot_beta_by_type(beta_df, 'all', return_fig=True, silent=True, on_lambda=self.on_lambda)
                        for fig in list_of_figs:
                            self.pdf.savefig(figure=fig, bbox_inches='tight')
                        self.plt.close('all')
                    elif part == 'predict_sex':
                        # get_sex() handles missing data issues; no need to validate files exist
                        include_probe_failure_percent = True if (self.tests['detection_poobah'] is True and 'detection_poobah' in self.order) else False
                        poobah_cutoff = (100 - self.__dict__.get('poobah_min_percent',80))
                        fig = methylcheck.get_sex(path, plot=True, save=False, on_lambda=self.on_lambda,
                            include_probe_failure_percent=include_probe_failure_percent, poobah_cutoff=poobah_cutoff,
                            return_fig=True) # custom_label={dict of values for samples keyed to sample IDs}
                        self.pdf.savefig(fig.fig) # seaborn relplots embed the matplotlib fig within them
                        self.plt.close()
                except Exception as e:
                    if self.debug:
                        LOGGER.error(f"Could not process {part}; {e}")
                        continue
                    else:
                        raise Exception(f"Could not process {part}; {e}")

            elif part in self.custom:
                table = self.custom[part]
                self.to_table(table['data'], col_names=table['col_names'],
                    row_names=table.get('row_names'), add_title=table['title'],
                    font_size=table.get('font_size',None))


        # and finally, appendix of log messages
        appendix_msgs = self.errors.getvalue()
        appendix_msgs = appendix_msgs.split('^^^')
        # drop the trailing \n at end of each message, if present, but preserve within-line \n line breaks
        appendix_msgs = [(msg[:-2] if msg[-2:] == '\n' else msg) for msg in appendix_msgs]
        appendix_msgs.insert(0, 'APPENDIX I: PROCESSING MESSAGES')
        total_rows = len(appendix_msgs)
        last_row = 0
        while last_row <= total_rows:
            rows_per_page = 26
            # pretest whether length will fit on a pgae
            while True:
                para_list = appendix_msgs[last_row : (last_row + rows_per_page)]
                para_lengths = [len(i.split('\n')) for i in para_list]
                extra_line_wraps = [max(len(self.textwrap.wrap(para, width=self.MAXWIDTH)) -1, 0) for para in para_list]
                #print((sum(para_lengths) + len(para_lengths) + sum(extra_line_wraps)))
                if (sum(para_lengths) + len(para_lengths) + sum(extra_line_wraps)) > self.MAXLINES:
                    rows_per_page -= 2
                    #print('---',rows_per_page)
                else:
                    break
                if rows_per_page <= 0:
                    break
            if rows_per_page <= 0:
                LOGGER.error("couldn't write log messages to appendix")
                break
            self.page_of_paragraphs(para_list, self.pdf, line_height='single')
            last_row += rows_per_page
        self.errors.close()


    def exec_summary(self):
        """QC exec summary
    	sample_name/ID
    	probe % failures
    	probe_failure pass
    	auto-qc result (only if present in kwargs passed in, otherwise omitted)
    	MDS pass
    	signal intensity pass
    	if any fails, fail it (so overall pass)
    table 2: meta
    	array type (detect from data)
    	number of samples (from data)
    	processing pipeline version number (passed in)
    	date processed (passed in)
    	avg probe failure rate
    	percent of samples that failed
    	any failures from 'control probes'
    	   reqs a way to capture warnings of data-off-chart
        """
        exec_summary_1 = []
        exec_summary_2 = [] # list of lists
        exec_summary_samples = [] # temp storage lookup
        pass

    def page_of_text(self, text, pdf):
        """text is a single big string of text, with whitespace for line breaks.
        https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.text.html (0,0) is lower left; (1,1) is upper right """
        print([len(i.split('\n')) for i in [text]])
        firstPage = self.plt.figure(figsize=(11,8.5))
        firstPage.clf()
        wrapped_txt = self.textwrap.fill(text, width=self.MAXWIDTH)
        #wrapped_txt_list = textwrap.wrap(txt, width=80)
        # next, identify paragraph breaks here as strings of writespace, and wrap each part separately? then move onto page.
        #firstPage.text(0.5,0.5,txt, size=12, ha="left") # transform=firstPage.transFigure
        firstPage.text(self.ORIGIN[0], self.ORIGIN[1], wrapped_txt, size=self.FONTSIZE)
        pdf.savefig()
        self.plt.close()


    def page_of_paragraphs(self, para_list, pdf, line_height='double'):
        """ https://matplotlib.org/3.1.1/api/_as_gen/matplotlib.pyplot.text.html (0,0) is lower left; (1,1) is upper right.
        this version estimates the size of each paragraph and moves the origin downward accordingly.
        tricky because the anchors are lower left, not upper left.

        ok if a paragraph contains whitespace line breaks, OR each para is one long line to be wrapped here.
        also - if a paragraph wraps, this accounts for it in total lines count"""
        #1 - finding the start points for each paragraph in page, counting backwards, so first paragraph is at top of page.
        para_lengths = [len(i.split('\n')) for i in para_list]
        extra_line_wraps = [max(len(self.textwrap.wrap(para, width=self.MAXWIDTH)) -1, 0) for para in para_list]
        if (sum(para_lengths) + len(para_lengths) + sum(extra_line_wraps)) > self.MAXLINES:
            raise ValueError(f"Text lines {(sum(para_lengths) + len(para_lengths) + sum(extra_line_wraps))} > {self.MAXLINES}, the max lines per PDF page allowed.")
        #print(para_lengths) -- ok if a paragraph contains whitespace line breaks, OR each para is one long line to be wrapped here.
        # assume each line height is 0.1/26 fraction of total page height

        # paragraph spacing
        if line_height == 'double':
            line_height = round((1 - 2*self.ORIGIN[1])/self.MAXLINES,3)
        elif line_height == 'single':
            line_height = round(0.5*(1 - 2*self.ORIGIN[1])/self.MAXLINES,3)
        else:
            pass
            # else, line_height is whatever int/float you pass in, for custom spacing. (as fraction of page, so use 0.01 to 0.03)

        paragraph_spacing = line_height # 1 line

        firstPage = self.plt.figure(figsize=(11,8.5))
        firstPage.clf()
        current_y_position = (1.0 - self.ORIGIN[1]) # working down page by subtracting
        extra_lines = [] # if something doesn't fit on page, include this in error message
        for nth, para in enumerate(para_list):
            # each para is assumed to be one paragraph of text with no whitespace formatting.
            wrapped_txt = self.textwrap.fill(para, width=self.MAXWIDTH)
            wrapped_addtl_line_count = max(len(self.textwrap.wrap(para, width=self.MAXWIDTH)) -1, 0)
            x_margin = self.ORIGIN[0]
            current_y_position -= (line_height * len(para.split('\n')))
            current_y_position -= (line_height * wrapped_addtl_line_count)
            if nth > 0:
                current_y_position -= paragraph_spacing
            if current_y_position < self.ORIGIN[1]:
                extra_lines.append(wrapped_txt)
                #print(current_y_position, wrapped_addtl_line_count, wrapped_txt)
            else:
                firstPage.text(x_margin, current_y_position, wrapped_txt, size=self.FONTSIZE)
        if extra_lines != []:
            print("WARNING: DID NOT FIT")
            for extra_line in extra_lines:
                print(extra_line)
                print("")
        pdf.savefig()
        self.plt.close()

    def to_table(self, list_of_lists, col_names, row_names=None, add_title='', font_size='auto'):
        """
        - embeds a table in a PDF page.
        - attempts to split long tables into multiple pages.
        - should warn if table is too wide to fit.
        - font_size:
          - auto: let matplotlib figure out the best size. This breaks if one column or value is soooooooper long.
          - truncate: IN FUTURE, chop off each value that is too long, to force it to fit on a page. Data loss.
          - INT: set the font for this table at this size (for control freaks)
          - None: default -- use font size 12."""

        # stringify numbers; replace None
        for idx, sub in enumerate(list_of_lists):
            list_of_lists[idx] = [('' if isinstance(i,type(None)) else str(i)) for i in sub]

        rows_per_page = 42 # appears to work for standard default PDF page size and font size.
        pages = math.ceil(len(list_of_lists)/rows_per_page)
        #paginate table
        for page in range(pages):
            #LOGGER.debug(f'page, cols: {len(col_names)}, rows: {len(list_of_lists)}')
            truncate = False
            if isinstance(font_size, str) and font_size.lower() == 'truncate':
                # overwrites list_of_lists so must happen before table is created
                truncate = True
                truncate_font_size = 12
                try:
                    max_field_length = 24
                    # default max length = 24 chars; if more than 4 columns, reduce by ((N-4)/20)*80 chars
                    if 4 < len(col_names) < 10:
                        max_field_length = int((24 - ((len(col_names)-4)/20) * 80))
                        if max_field_length <= 4: # 9 col
                            truncate_font_size = 6
                            max_field_length += 8
                        elif max_field_length <= 8: # 8 col
                            truncate_font_size = 7
                            max_field_length += 3
                        elif max_field_length <= 12: # 7 col
                            truncate_font_size = 7
                            max_field_length += 1
                        elif max_field_length <= 16: # 6 col
                            truncate_font_size = 7
                            #max_field_length += 1
                        elif max_field_length <= 20: # 5 col
                            truncate_font_size = 7
                            #max_field_length += 0
                        elif max_field_length <= 24: # 4 col
                            truncate_font_size = 8
                            #max_field_length += 0
                    elif len(col_names) >= 10:
                        LOGGER.error(f"Could not truncate long fields in custom table because there were 10 or more columns.")
                        truncate = False
                    if truncate is True:
                        for row_num, row in enumerate(list_of_lists.copy()):
                            for field_num,field in enumerate(row):
                                if len(str(field)) > max_field_length:
                                    list_of_lists[row_num][field_num] = str(field)[:max_field_length] + '..'
                                    if self.debug:
                                        LOGGER.info(f"{field} truncated to {field[:max_field_length]}..")
                        for field_num,field in enumerate(col_names):
                            if len(str(field)) > max_field_length:
                                col_names[field_num] = str(field)[:max_field_length] + '..'
                                if self.debug:
                                    LOGGER.info(f"column name {field} truncated to {field[:max_field_length]}..")
                except Exception as e:
                    truncate = False
                    LOGGER.error(f"Could not truncate long fields in custom table: {e}")
            page_data = list_of_lists[(page * rows_per_page) : (page + 1) * rows_per_page]
            fig, ax = self.plt.subplots(1, figsize=(8,10.5))
            ax.axis('off')
            matplotlib_table = ax.table(
                cellText=page_data,
                colLabels=col_names,
                loc='center', # if len(page_data) >= rows_per_page else 'top',
                edges='open',
                colLoc='right',
                )

            if font_size is None:
                matplotlib_table.auto_set_font_size(False)
                matplotlib_table.set_fontsize(12)
            elif isinstance(font_size, int):
                matplotlib_table.auto_set_font_size(False)
                matplotlib_table.set_fontsize(font_size)
            elif ( (isinstance(font_size, str) and font_size.lower() == 'auto') or
                   (isinstance(font_size, str) and font_size.lower() == 'truncate' and truncate is False) ): #default shrinks to fit page
                matplotlib_table.auto_set_font_size(True)
                scale_ratio = 1.2 # use smaller page margins
                matplotlib_table.scale(scale_ratio, scale_ratio)
            elif truncate is True:
                matplotlib_table.auto_set_column_width(col=list(range(len(col_names)))) # Provide integer list of columns to adjust
                matplotlib_table.auto_set_font_size(False)
                matplotlib_table.set_fontsize(truncate_font_size)
                scale_ratio = 1.2
                matplotlib_table.scale(scale_ratio, scale_ratio)
            if page == 0:
                self.plt.title(add_title, y=1.1) #pad=20) # -- placement is off
            #fig.set_size_inches(8,10.5)
            self.pdf.savefig(fig)
            self.plt.close()
